"""
Two-Point Correlator Algorithms (:mod:`~triumvirate._twopt`)
==========================================================================

Declare and interface with two-point correlator algorithms.

"""
from cython.operator cimport dereference as deref
from libc.stdlib cimport free, malloc
from libcpp.string cimport string

import numpy as np
cimport numpy as np

from triumvirate._particles cimport CppParticleCatalogue, _ParticleCatalogue
from triumvirate.dataobjs cimport (
    Binning, CppBinning,
    LineOfSight,
    PowspecMeasurements, TwoPCFMeasurements, TwoPCFWindowMeasurements
)
from triumvirate.parameters cimport CppParameterSet, ParameterSet


cdef extern from "include/twopt.hpp":
    # --------------------------------------------------------------------
    # Normalisation
    # --------------------------------------------------------------------

    double calc_powspec_normalisation_from_particles_cpp \
        "trv::calc_powspec_normalisation_from_particles" (
            CppParticleCatalogue& particles,
            double alpha
        ) except +

    double calc_powspec_normalisation_from_mesh_cpp \
        "trv::calc_powspec_normalisation_from_mesh" (
            CppParticleCatalogue& particles,
            CppParameterSet& params,
            double alpha
        )

    double calc_powspec_normalisation_from_meshes_cpp \
        "trv::calc_powspec_normalisation_from_meshes" (
            CppParticleCatalogue& particles_data,
            CppParticleCatalogue& particles_rand,
            CppParameterSet& params,
            double alpha
        )

    double calc_powspec_normalisation_from_meshes_cpp \
        "trv::calc_powspec_normalisation_from_meshes" (
            CppParticleCatalogue& particles_data,
            CppParticleCatalogue& particles_rand,
            CppParameterSet& params,
            double alpha,
            double padding,
            double cellsize,
            string assignment
        )


    # --------------------------------------------------------------------
    # Full correlator
    # --------------------------------------------------------------------

    PowspecMeasurements compute_powspec_cpp "trv::compute_powspec" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& kbinning,
        double norm_factor
    )

    TwoPCFMeasurements compute_corrfunc_cpp "trv::compute_corrfunc" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& rbinning,
        double norm_factor
    )

    PowspecMeasurements compute_powspec_in_gpp_box_cpp \
        "trv::compute_powspec_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& kbinning,
            double norm_factor
        )

    TwoPCFMeasurements compute_corrfunc_in_gpp_box_cpp \
        "trv::compute_corrfunc_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& rbinning,
            double norm_factor
        )

    TwoPCFWindowMeasurements compute_corrfunc_window_cpp \
        "trv::compute_corrfunc_window" (
            CppParticleCatalogue& catalogue_rand,
            LineOfSight* los_rand,
            CppParameterSet& params,
            CppBinning& rbinning,
            double alpha,
            double norm_factor
        )


def _calc_powspec_normalisation_from_particles(
        _ParticleCatalogue particles not None, double alpha
    ):
    return calc_powspec_normalisation_from_particles_cpp(
        deref(particles.thisptr), alpha
    )


def _calc_powspec_normalisation_from_mesh(
        _ParticleCatalogue particles not None,
        ParameterSet params not None,
        double alpha
    ):
    return calc_powspec_normalisation_from_mesh_cpp(
        deref(particles.thisptr), deref(params.thisptr), alpha
    )


def _calc_powspec_normalisation_from_meshes(
        _ParticleCatalogue particles_data not None,
        _ParticleCatalogue particles_rand not None,
        ParameterSet params not None,
        double alpha,
        padding=None, cellsize=None, assignment=None
    ):
    if None in [padding, cellsize, assignment]:
        return calc_powspec_normalisation_from_meshes_cpp(
            deref(particles_data.thisptr), deref(particles_rand.thisptr),
            deref(params.thisptr), alpha
        )
    else:  # STYLE: non-Pythonic use of `else`
        if not (
            isinstance(padding, float)
            and isinstance(cellsize, float)
            and isinstance(assignment, str)
        ):
            raise TypeError(
                "`padding`, `cellsize` and `assignment` must be "
                "of type float, float and str: received {}, {} and {}."
                .format(type(padding), type(cellsize), type(assignment))
            )
        return calc_powspec_normalisation_from_meshes_cpp(
            deref(particles_data.thisptr), deref(particles_rand.thisptr),
            deref(params.thisptr), alpha,
            padding, cellsize, assignment.encode('utf-8')
        )


def _compute_powspec(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef PowspecMeasurements results
    results = compute_powspec_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'kbin': np.array(results.kbin),
        'keff': np.array(results.keff),
        'nmodes': np.array(results.nmodes),
        'pk_raw': np.array(results.pk_raw),
        'pk_shot': np.array(results.pk_shot),
    }


def _compute_corrfunc(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef TwoPCFMeasurements results
    results = compute_corrfunc_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'rbin': np.array(results.rbin),
        'reff': np.array(results.reff),
        'npairs': np.array(results.npairs),
        'xi': np.array(results.xi),
    }


def _compute_powspec_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    cdef PowspecMeasurements results
    results = compute_powspec_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    return {
        'kbin': np.array(results.kbin),
        'keff': np.array(results.keff),
        'nmodes': np.array(results.nmodes),
        'pk_raw': np.array(results.pk_raw),
        'pk_shot': np.array(results.pk_shot),
    }


def _compute_corrfunc_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    cdef TwoPCFMeasurements results
    results = compute_corrfunc_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    return {
        'rbin': np.array(results.rbin),
        'reff': np.array(results.reff),
        'npairs': np.array(results.npairs),
        'xi': np.array(results.xi),
    }


def _compute_corrfunc_window(
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double alpha,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef TwoPCFWindowMeasurements results
    results = compute_corrfunc_window_cpp(
        deref(catalogue_rand.thisptr), los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        alpha, norm_factor
    )

    free(los_rand_cpp)

    return {
        'rbin': np.array(results.rbin),
        'reff': np.array(results.reff),
        'npairs': np.array(results.npairs),
        'xi': np.array(results.xi),
    }



"""
Window Convolution (:mod:`~triumvirate.winconv`)
==========================================================================

Perform window convolution of three-point statistics.

"""
from typing import NamedTuple

import numpy as np
from scipy.interpolate import interp1d, RectBivariateSpline

from triumvirate._bihankel import (
    transform_bispec_to_3pcf,
    transform_3pcf_to_bispec,
)


# Default convolution formula from Sugiyama et al. (2019) [arXiv:1803.02132].
_FORMULAE_SUGIYAMA19 = {
    '000': [
        ('000', '000', 1),
        ('000', 'ic', -1),
        ('110', '110', 1./3),
    ],
    '110': [
        ('000', '110', 1),
        ('110', '000', 1),
        ('110', 'ic', -1),
    ],
    '202': [
        ('000', '202', 1),
        ('202', '000', 1),
        ('202', 'ic', -1),
        ('110', '112', 1./3),
        ('112', '110', 1./3),
    ],
    '112': [
        ('000', '112', 1),
        ('112', '000', 1),
        ('112', 'ic', -1),
        ('022', '110', 2./5),
        ('202', '110', 2./5),
        ('110', '022', 2./5),
        ('110', '202', 2./5),
    ],
}


class _WConvTerm(NamedTuple):
    r"""Window convolution term as a tuple of its three factors.

    For example, ``term = _WConvTerm('000', 'ic', -1)``, with
    parameters/attributes/factors ``term.deg_Q = '000'``,
    ``term.deg_zeta = 'ic'`` and ``term.coeff = -1``, corresponds to the
    term :math:`- 1 \cdot Q_{000} \zeta_{\mathrm{ic}}`.

    Attributes
    ----------
    deg_Q : str
        Window function multipole degree.
    deg_zeta : str
        Unwindowed 3-point correlation funciton multipole degree.
    coeff : float
        Numerical factor for the convolution term.

    """
    deg_Q: str
    deg_zeta: str
    coeff: float


class _WConvFormulae:
    """Window convolution formulae.

    The full formulae are encoded as a dictionary, where each key
    corresponds to a windowed 3-point correlation function (3PCF)
    multipole, and each value is a sequence of (named) tuples each
    corresponding to a window convolution term.

    Parameters
    ----------
    formulae : dict of {str: sequence of tuples of (str, str, float)}
        Window convolution formulae.

    Attributes
    ----------
    multipoles_Q : list of str
        List of degrees of the required window function multipoles.
    multipoles_zeta : list of str
        List of degrees of the required 3PCF multipoles.

    """

    def __init__(self, formulae):

        self._formulae = {
            multipole: [_WConvTerm(*term) for term in formula]
            for multipole, formula in formulae.items()
        }

        self.multipoles = formulae.keys()

        self.multipoles_Q = np.unique([
            term.deg_Q
            for formula in self._formulae.values()
            for term in formula
        ])
        self.multipoles_zeta = np.unique([
            term.deg_zeta
            for formula in self._formulae.values()
            for term in formula
        ])

    def __getitem__(self, multipole):
        """Get the formula for the window convolution of a specific
        3PCF multipole.

        Parameters
        ----------
        multipole : str
            Degrees of the specified windowed 3PCF multipole.

        Returns
        -------
        list of :class:`~triumvirate.winconv._WConvTerm`
            List of window convolution terms for `multipole`.

        """
        return self._formulae[multipole]


def wconv_3pcf_diag(formulae, Q, zeta, r_common, r_Q=None, r_zeta=None):
    r"""Convolve multipoles of the 3-point correlation function (3PCF) and
    the window function in diagonal form using specified formulae.

    Parameters
    ----------
    formulae : dict of {str: sequence of (str, str, float)} or 'sugiyama+19'
        Window convolution formulae encoded as a dictionary where for
        each convolved-multipole degree (the key), the value is
        a sequence of tuples each corresponding to a window convolution
        term including the numerical coefficient.

        For example,

        .. code-block::

            formulae = {
                '000': [
                    ('000', '000', 1),
                    ('000', 'ic', -1),
                    ('110', '110', 1./3),
                ],
            }

        computes only the '000' monopole

        .. math::

            \tilde{\zeta}_{000} =
                1 \cdot Q_{000} \zeta_{000}
                + (-1) \cdot Q_{000} \zeta_{\mathrm{ic}}
                + \frac{1}{3} \cdot Q_{110} \zeta_{110} \,.

        If a named formula (e.g. 'sugiyama19') is passed, the window
        convolution formula follows that prescription
        (e.g. :cite:t:`Sugiyama:2019`).

    Q : dict of {str: 2-d array of float}
        Window function multipole samples (each key is a multipole).
    zeta : dict of {str: 2-d array of float}
        Unwindowed 3PCF multipole samples (each key is a multipole).
    r_common : 1-d array of float
        Common separation sample points for both `Q` and `zeta`.
        If `r_Q` or `r_zeta` is provided, all multipole samples are
        re-interpolated over `r_common`; otherwise all multipole
        sample points are assumed to be `r_common`.
    r_Q : 1-d array of float, optional
        Separation sample points for `Q`.  If `None` (default),
        `r_common` is assumed instead.
    r_zeta : 1-d array of float, optional
        Separation sample points for `zeta`.  If `None` (default),
        `r_common` is assumed instead.

    Returns
    -------
    zeta_wconv : dict of {str: :class:`numpy.ndarray`}
        Window-convolved 3PCF samples.

    """
    # Fetch named formulae as a dictionary.
    if isinstance(formulae, str) and formulae.lower() == 'sugiyama+19':
        formulae = _FORMULAE_SUGIYAMA19

    # Convert dictionary to formular object.
    if not isinstance(formulae, _WConvFormulae):
        formulae = _WConvFormulae(formulae)

    # Coerce onto the same coordinate samples.
    if r_Q is None:
        Q_samples = Q
    else:
        r_Q = r_Q if r_Q is not None else r_common
        Q_samples = {
            multipole: interp1d(r_Q, Q[multipole])(r_common)
            for multipole in Q.keys()
        }

    if r_zeta is None:
        zeta_samples = zeta
    else:
        r_zeta = r_zeta if r_zeta is not None else r_common
        zeta_samples = {
            multipole: interp1d(r_zeta, zeta[multipole])(r_common)
            for multipole in zeta.keys()
        }

    # Perform convolution as multiplication.
    zeta_conv = {}
    for multipole in formulae.multipoles:
        zeta_tilde = 0.  # technically not a scalar
        for term in formulae[multipole]:
            zeta_tilde += term.coeff \
                * Q_samples[term.deg_Q] * zeta_samples[term.deg_zeta]
        zeta_conv[multipole] = zeta_tilde

    return zeta_conv


def wconv_3pcf(formulae, Q, zeta, r1_common, r2_common,
               r1_Q=None, r2_Q=None, r1_zeta=None, r2_zeta=None):
    r"""Convolve multipoles of the 3-point correlation function (3PCF) and
    the window function using specified formulae.

    Parameters
    ----------
    formulae : dict of {str: sequence of (str, str, float)} or 'sugiyama+19'
        Window convolution formulae encoded as a dictionary where for
        each convolved-multipole degree (the key), the value is
        a sequence of tuples each corresponding to a window convolution
        term including the numerical coefficient.

        For example,

        .. code-block::

            formulae = {
                '000': [
                    ('000', '000', 1),
                    ('000', 'ic', -1),
                    ('110', '110', 1./3),
                ],
            }

        computes only the '000' monopole

        .. math::

            \tilde{\zeta}_{000} =
                1 \cdot Q_{000} \zeta_{000}
                + (-1) \cdot Q_{000} \zeta_{\mathrm{ic}}
                + \frac{1}{3} \cdot Q_{110} \zeta_{110} \,.

        If a named formula (e.g. 'sugiyama19') is passed, the window
        convolution formula follows that prescription
        (e.g. :cite:t:`Sugiyama:2019`).

    Q : dict of {str: 2-d array of float}
        Window function multipole samples (each key is a multipole).
    zeta : dict of {str: 2-d array of float}
        Unwindowed 3PCF multipole samples (each key is a multipole).
    r1_common, r2_common : 1-d array of float
        Common separation sample points for both `Q` and `zeta`.
        If `r1_Q`/`r1_zeta` or `r2_Q`/`r2_zeta` is provided,
        all multipole samples are re-interpolated over `r1_common`
        and `r2_common`; otherwise all multipole sample points are assumed
        to be `r1_common` and `r2_common`.
    r1_Q, r2_Q : 1-d array of float, optional
        Separation sample points for `Q`.  If `None` (default),
        `r1_common` and `r2_common` are assumed instead.
    r1_zeta, r2_zeta : 1-d array of float, optional
        Separation sample points for `zeta`.  If `None` (default),
        `r1_common` and `r2_common` are assumed instead.

    Returns
    -------
    zeta_wconv : dict of {str: :class:`numpy.ndarray`}
        Window-convolved 3PCF samples

    """
    # Fetch named formulae as a dictionary.
    if isinstance(formulae, str) and formulae.lower() == 'sugiyama+19':
        formulae = _FORMULAE_SUGIYAMA19

    # Convert dictionary to formular object.
    if not isinstance(formulae, _WConvFormulae):
        formulae = _WConvFormulae(formulae)

    # Coerce onto the same coordinate samples.
    if r1_Q is None and r2_Q is None:
        Q_samples = Q
    else:
        r1_Q = r1_Q if r1_Q is not None else r1_common
        r2_Q = r2_Q if r2_Q is not None else r2_common
        Q_samples = {
            multipole: RectBivariateSpline(
                r1_Q, r2_Q, Q[multipole]
            )(r1_common, r2_common)
            for multipole in Q.keys()
        }

    if r1_zeta is None and r2_zeta is None:
        zeta_samples = zeta
    else:
        r1_zeta = r1_zeta if r1_zeta is not None else r1_common
        r2_zeta = r2_zeta if r2_zeta is not None else r2_common
        zeta_samples = {
            multipole: RectBivariateSpline(
                r1_zeta, r2_zeta, zeta[multipole]
            )(r1_common, r2_common)
            for multipole in zeta.keys()
        }

    # Perform convolution as multiplication.
    zeta_conv = {}
    for multipole in formulae.multipoles:
        zeta_tilde = 0.  # technically not a scalar
        for term in formulae[multipole]:
            zeta_tilde += term.coeff \
                * Q_samples[term.deg_Q] * zeta_samples[term.deg_zeta]
        zeta_conv[multipole] = zeta_tilde

    return zeta_conv


def wconv_bispec(formulae, Q, bk, k, k_out, r_common, r_Q=None,
                 transform_kwargs=None):
    r"""Convolve multipoles of the bispectrum and the window function
    via configuration space using specified formulae.

    Parameters
    ----------
    formulae : dict of {str: sequence of (str, str, float)} or 'sugiyama+19'
        See :func:`~triumvirate.winconv.wconv_3pcf`.
    Q : dict of {str: 2-d array of float}
        Configuration-space window function multipole samples
        (each key is a multipole).
    bk : dict of {str: 2-d array of float}
        Unwindowed bispectrum multipole samples (each key is a multipole).
    k : 1-d array of float
        Unwindowed bispectrum wavenumber samples.
    k_out : 1-d array of float
        Output bispectrum wavenumber samples.
    r_common : 1-d array of float
        Common separation sample points for both `Q` and 3PCF :math:`zeta`
        (transformed from `bk`).  If `r_Q` is provided, all multipole
        samples are re-interpolated over `r_common`; otherwise all
        multipole sample points are assumed to be `r_common`.
    r_Q : 1-d array of float, optional
        Separation sample points for `Q`.  If `None` (default),
        `r_common` is assumed instead.
    transform_kwargs : dict, optional
        Keyword arguments to be passed to
        :class:`~triumvirate.bihankel.transform_bispec_to_3pcf` and
        :class:`~triumvirate.bihankel.transform_3pcf_to_bispec` for
        transforming to/from configuration space.  If `None` (default),
        the FFTLog sample number is set to 1024 and no extrapolation
        is performed.

    Returns
    -------
    bk_wconv : dict of {str: :class:`numpy.ndarray`}
        Window-convolved bispectrum samples.


    .. attention::

        This function assumes the wavenumber samples :math:`k_{1,2}` in
        the bispectrum :math:`B_{\ell_1 \ell_2 L}(k_1, k_2)` are the same.

    """
    if transform_kwargs is None:
        transform_kwargs = {}
    n_fftlog = transform_kwargs.get('n_fftlog', 1024)

    zeta = {}
    for multipole, bk_in in bk.items():
        ell1, ell2 = int(multipole[0]), int(multipole[1])
        zeta[multipole] = transform_bispec_to_3pcf(
            ell1, ell2, k, bk_in, r_common, n_fftlog,
            **transform_kwargs
        )['zeta']

    zeta_wconv = wconv_3pcf(
        formulae, Q, zeta, r_common, r_common, r1_Q=r_Q, r2_Q=r_Q
    )

    bk_wconv = {}
    for multipole, zeta_in in zeta_wconv.items():
        ell1, ell2 = int(multipole[0]), int(multipole[1])
        bk_wconv[multipole] = transform_3pcf_to_bispec(
            ell1, ell2, r_common, zeta_in, k_out, n_fftlog
        )

    return bk_wconv



"""Configure `pytest`.

"""
import inspect
import warnings
from pathlib import Path

import numpy as np
import pytest

from triumvirate.catalogue import ParticleCatalogue
from triumvirate.dataobjs import Binning
from triumvirate.logger import setup_logger
from triumvirate.parameters import ParameterSet, fetch_paramset_template


# ========================================================================
# Paths
# ========================================================================

conftest_file = inspect.getframeinfo(inspect.currentframe()).filename


# ========================================================================
# Configuration
# ========================================================================

def pytest_addoption(parser):
    # Add '--runslow' option for slow test functions.
    parser.addoption(
        '--runslow', action='store_true', default=False, help="run slow tests"
    )


def pytest_configure(config):
    # Add 'slow' marker for tests to run only when running slow tests
    # is enabled.
    config.addinivalue_line('markers', "slow: mark test as slow to run")

    # Add 'no_capture' marker for tests to run only when output capturing
    # is disabled.
    config.addinivalue_line(
        'markers',
        "no_capture: mark test to run only when output capturing is disabled"
    )


def pytest_collection_modifyitems(config, items):
    # Skip tests marked with 'no_capture' when capturing is not disabled.
    if not config.getoption('--runslow'):
        skip_slow = pytest.mark.skip(
            reason="need '--runslow' option to run slow tests"
        )
        for item in items:
            if 'slow' in item.keywords:
                item.add_marker(skip_slow)
    if config.getoption('--capture') != 'no':
        skip_captured = pytest.mark.skip(
            reason="need '--capture=no' or '-s' option to run"
        )
        for item in items:
            if 'no_capture' in item.keywords:
                item.add_marker(skip_captured)


# ========================================================================
# Fixtures
# ========================================================================

@pytest.fixture(scope='session')
def test_dir():
    return Path(conftest_file).parent.resolve()


@pytest.fixture(scope='session')
def test_input_dir(test_dir):
    return test_dir/"test_input"


@pytest.fixture(scope='session')
def test_output_dir(test_dir):
    return test_dir/"test_output"


@pytest.fixture(scope='session')
def test_param_dir(test_input_dir):
    return test_input_dir/"params"


@pytest.fixture(scope='session')
def test_ctlg_dir(test_input_dir):
    return test_input_dir/"ctlgs"


@pytest.fixture(scope='session')
def test_stats_dir(test_input_dir):
    return test_input_dir/"stats"


@pytest.fixture
def test_logger():
    return setup_logger()


@pytest.fixture
def valid_paramset():
    param_dict = fetch_paramset_template('dict')

    for ax_name in ['x', 'y', 'z']:
        param_dict['boxsize'][ax_name] = 1000.
        param_dict['ngrid'][ax_name] = 64.

    param_dict.update({
        'statistic_type': 'bispec',
        'degrees': {'ell1': 0, 'ell2': 0, 'ELL': 0},
        'range': [0.005, 0.105],
        'num_bins': 10,
        'verbose': 60,
    })

    return ParameterSet(param_dict=param_dict)


@pytest.fixture
def test_paramset(test_param_dir):
    return ParameterSet(param_filepath=test_param_dir/"test_params.yml")


@pytest.fixture
def test_binning_fourier():
    return Binning('fourier', 'lin', bin_min=0.005, bin_max=0.105, num_bins=4)


@pytest.fixture
def test_binning_config():
    return Binning('config', 'lin', bin_min=50., bin_max=150., num_bins=4)


@pytest.fixture
def test_catalogue_properties():
    return {
        'separation': 100.,
        'boxsize': 1.e3,
        'nparticle': 3,
        'contrast': 1.e4,
    }


@pytest.fixture
def test_data_catalogue(test_ctlg_dir, test_logger, test_catalogue_properties):

    warnings.filterwarnings('ignore', message=".*field is not provided.*")

    if (test_ctlg_dir/"test_data_catalogue.txt").exists():
        return ParticleCatalogue.read_from_file(
            test_ctlg_dir/"test_data_catalogue.txt",
            names=['x', 'y', 'z', 'nz'],
            logger=test_logger
        )

    R = test_catalogue_properties['separation']
    L = test_catalogue_properties['boxsize']
    N = test_catalogue_properties['nparticle']

    x = [-R/2., 0., R/2.]
    y = [-R/2./np.sqrt(3), R/np.sqrt(3), -R/2./np.sqrt(3)]
    z = [0., 0., 0.]
    nz = N / L**3

    return ParticleCatalogue(x, y, z, nz=nz, logger=test_logger)


@pytest.fixture
def test_rand_catalogue(test_ctlg_dir, test_logger, test_catalogue_properties):

    warnings.filterwarnings('ignore', message=".*field is not provided.*")

    if (test_ctlg_dir/"test_rand_catalogue.txt").exists():
        return ParticleCatalogue.read_from_file(
            test_ctlg_dir/"test_rand_catalogue.txt",
            names=['x', 'y', 'z', 'nz'],
            logger=test_logger
        )

    alpha = test_catalogue_properties['contrast']
    L = test_catalogue_properties['boxsize']
    N = test_catalogue_properties['nparticle']

    generator = np.random.default_rng(seed=42)

    x, y, z = generator.uniform(-L/2., L/2., size=(3, int(alpha*N)))
    nz = N / L**3

    return ParticleCatalogue(x, y, z, nz=nz, logger=test_logger)


@pytest.fixture
def test_uniform_catalogue(test_ctlg_dir, test_logger,
                           test_catalogue_properties):

    warnings.filterwarnings('ignore', message=".*field is not provided.*")

    if (test_ctlg_dir/"test_uniform_catalogue.txt").exists():
        return ParticleCatalogue.read_from_file(
            test_ctlg_dir/"test_uniform_catalogue.txt",
            names=['x', 'y', 'z', 'nz'],
            logger=test_logger
        )

    alpha = test_catalogue_properties['contrast'] / 10.
    L = test_catalogue_properties['boxsize']
    N = test_catalogue_properties['nparticle']

    generator = np.random.default_rng(seed=69)

    x, y, z = generator.uniform(-L/2., L/2., size=(3, int(alpha*N)))
    nz = N / L**3

    return ParticleCatalogue(x, y, z, nz=nz, logger=test_logger)



"""
Catalogue (:mod:`~triumvirate.catalogue`)
==========================================================================

Handle catalogue I/O and processing.

.. autosummary::
    MissingValueError
    DefaultValueWarning
    ParticleCatalogue

"""
import warnings

import numpy as np
from astropy.table import Table

from triumvirate._particles import _ParticleCatalogue

try:
    from nbodykit.source.catalog import (
        ArrayCatalog,
        BinaryCatalog,
        CSVCatalog,
        FITSCatalog,
        HDFCatalog,
    )
    _nbkt_imported = True
except Exception:
    _nbkt_imported = False


class MissingValueError(ValueError):
    """Value error raised when a mandatory field is missing/empty in
    a catalogue.

    """
    pass


class DefaultValueWarning(UserWarning):
    """Warning issued when values of a field are not provided and set
    to default.

    """
    pass


class ParticleCatalogue:
    """Catalogue holding particle coordinates, weights and
    redshift-dependent mean number densities.

    Parameters
    ----------
    x, y, z : 1-d array of float
        Cartesian coordinates of particles.  `x`, `y` and `z` must have
        the same length.
    nz : (1-d array of) float, optional
        Redshift-dependent mean number density of
        (defaults is `None`).  If an array, it must be of the same length
        as `x`, `y` and `z`.
    ws, wc : (1-d array of) float, optional
        Sample weights and clustering weights of particles (defaults
        are 1.).  If an array, it must be of the same length as
        `x`, `y` and `z`.
    logger : :class:`logging.Logger`, optional
        Program logger (default is `None`).

    Attributes
    ----------
    bounds : dict of {str: tuple of (float, float)}
        Particle coordinate bounds.
    ntotal : int
        Total particle number.
    wtotal : float
        Total particle overall weight.
    wstotal : float
        Total particle sample weight.


    .. attention::

        There are two types of weights: sample weights ``ws``
        (e.g. completeness weights) and clustering weights ``wc``
        (e.g. Feldman--Kaiser--Peacock weights). The overall weight
        is the product of the two for each particle.

        Note the naming convention above: in particular, ``wc`` is not
        the completeness weight (which is a component of ``ws`` instead).

    """

    def __init__(self, x, y, z, nz=None, ws=1., wc=1., logger=None):

        self._logger = logger

        # Initialise particle data (with 'astropy').
        if nz is None:
            warnings.warn(
                "Catalogue 'nz' field is None and thus set to zero, "
                "which may raise errors in some computations.",
                category=DefaultValueWarning
            )
            nz = 0.

        self._pdata = Table()
        self._pdata['x'] = x
        self._pdata['y'] = y
        self._pdata['z'] = z
        self._pdata['nz'] = nz
        self._pdata['ws'] = ws
        self._pdata['wc'] = wc

        if _nbkt_imported:
            self._pdata = ArrayCatalog(self._pdata)
            self._backend = 'nbodykit'
        else:
            self._backend = 'astropy'

        self._source = 'extdata:{}'.format(id(self._pdata))

        # Compute catalogue properties.
        self._calc_bounds(init=True)

        self.ntotal = len(self)
        self.wtotal = self._compute(
            (self._pdata['ws'] * self._pdata['wc']).sum()
        )
        self.wstotal = self._compute(self._pdata['ws'].sum())

        if self._logger:
            self._logger.info(
                "Catalogue initialised: "
                "ntotal = %d, wtotal = %.3f, wstotal = %.3f (%s).",
                self.ntotal, self.wtotal, self.wstotal, self
            )

    @classmethod
    def read_from_file(cls, filepath, reader='astropy', names=None,
                       format='ascii.no_header', table_kwargs=None,
                       file_kwargs=None, name_mapping=None, logger=None):
        """Read particle data from file.

        Parameters
        ----------
        filepath : str or :class:`pathlib.Path`
            Catalogue file path.
        reader : {'astropy', 'nbodykit'}, optional
            If 'astropy' (default), |Table| is used for reading in the
            catalogue file; else if 'nbodykit', child classes of
            |FileCatalogBase| is used (if :mod:`nbodykit` is available).
        names : sequence of str, list of tuple or :class:`numpy.dtype`, \
                optional
            Catalogue file field names or data types.  If `None`
            (default), the header in the file may be used to infer the
            field names or data types.  See the note below for
            more details.
        format : str, optional
            File format specifier (default is 'ascii.no_header' for
            default `reader` value 'astropy').  See the note below for
            more details.
        name_mapping : dict of {str: str}, optional
            Mapping between any of the default column names 'x', 'y', 'z',
            'nz', 'ws' and 'wc' (keys) and the corresponding field names
            (values) in `names` (default is `None`).
        table_kwargs : dict, optional
            Keyword arguments to be passed to
            :attr:`astropy.table.Table.read` (default is `None`).
            Used only when ``reader='astropy'``.
        file_kwargs : dict, optional
            Keyword arguments to be passed to child classes of
            :class:`nbodykit.source.catalog.file.FileCatalogBase` (default
            is `None`). Used only when ``reader='nbodykit'``.  See also
            the hint below.
        logger : :class:`logging.Logger`, optional
            Program logger (default is `None`).

        Examples
        --------
        >>> ParticleCatalogue.read_from_file(
        ...     filepath,
        ...     reader='astropy',
        ...     format='fits',
        ...     # Match original data column names to the default names.
        ...     name_mapping={
        ...         'x': 'X', 'y': 'Y', 'z': 'Z',
        ...         'nz': 'NZ', 'ws': 'WEIGHT_SYS', 'wc': 'WEIGHT_FKP',
        ...     }
        ... )
        >>> ParticleCatalogue.read_from_file(
        ...     filepath,
        ...     reader='nbodykit',
        ...     format='hdf5',
        ...     # `root` keyword argument is passed to `HDFCatalog`.
        ...     file_kwargs={root='particles'}
        ... )


        .. admonition:: `format` and `names` arguments
            :class: dropdown

            For ``reader='astropy'``, supported `format` can be found in
            `'Built-In Table Readers/Writers'`_, and `names` correspond to
            the ``names`` keyword argument (sequence of str) in
            :attr:`astropy.table.Table.read` for a subset of formats.
            See |Table| for more details including appropriate
            `table_kwargs`.

            For ``reader='nbodykit'``, supported `format` and the
            argument corresponding to `names` are---

            - ``'text'`` and ``names`` (sequence of str): plain-text files
              read in by |CSVCatalog|;
            - ``'binary'`` and ``dtype`` (list of tuple or
              :class:`numpy.dtype`): binary files read in
              by |BinaryCatalog|;
            - ``'hdf5'`` and not applicable: HDF files read in
              by |HDFCatalog| (pass `file_kwargs` as appropriate).

            See `'Reading Catalogs from Disk'`_ for more details.


        .. hint::

            If any of the 'nz', 'ws' and 'wc' columns are not provided,
            these columns are initialised with the default values in
            :class:`~triumvirate.catalogue.ParticleCatalogue`.


        .. |Table| replace:: :class:`astropy.table.Table`

        .. |FileCatalogBase| replace:: \
            :class:`nbodykit.source.catalog.file.FileCatalogBase`

        .. |CSVCatalog| replace:: \
            :class:`nbodykit.source.catalog.file.CSVCatalog`

        .. |FITSCatalog| replace:: \
            :class:`nbodykit.source.catalog.file.FITSCatalog`

        .. |BinaryCatalog| replace::  \
            :class:`nbodykit.source.catalog.file.BinaryCatalog`

        .. |HDFCatalog| replace:: \
            :class:`nbodykit.source.catalog.file.HDFCatalog`

        .. _'Built-In Table Readers/Writers': \
            https://docs.astropy.org/en/latest/io/unified.html
            #built-in-table-readers-writers

        .. _'Reading Catalogs from Disk': \
            https://nbodykit.readthedocs.io/en/latest/catalogs/reading.html

        """
        self = object.__new__(cls)

        self._logger = logger
        self._source = f'extfile:{filepath}'

        if reader.lower() == 'nbodykit':
            if _nbkt_imported:
                self._backend = 'nbodykit'
            else:
                warnings.warn(
                    "'astropy' is used for catalogue I/O as "
                    "'nbodykit' is unavailable",
                    category=RuntimeWarning
                )
                self._backend = 'astropy'
        elif reader.lower() == 'astropy':
            self._backend = 'astropy'
        else:
            raise ValueError(
                f"Unsupported catalogue file reader: {reader}. "
                "Possible options: {'astropy', 'nbodykit'}."
            )

        # Initialise particle data.
        if self._backend == 'nbodykit':
            if file_kwargs is None:
                file_kwargs = {}

            filepath = str(filepath)
            if format == 'text':
                self._pdata = CSVCatalog(filepath, names, **file_kwargs)
            elif format == 'fits':
                raise ValueError(
                    "'nbodykit' FITS reader is broken. "
                    "Use 'astropy' FITS reader instead."
                )
            elif format == 'binary':
                self._pdata = \
                    BinaryCatalog(filepath, dtype=names, **file_kwargs)
            elif format == 'hdf5':
                self._pdata = HDFCatalog(filepath, **file_kwargs)
            else:
                raise ValueError(
                    "Unsupported `format` for ``reader='nbodykit'``: {}."
                    .format(filepath)
                )

            if name_mapping:
                for name_, name_alt_ in name_mapping.items():
                    self._pdata[name_] = self._pdata[name_alt_]
                    try:
                        del self._pdata[name_alt_]
                    except ValueError:
                        pass

        if self._backend == 'astropy':
            if table_kwargs is None:
                table_kwargs = {}
            if format.startswith('ascii'):
                table_kwargs.update(names=names)

            self._pdata = Table.read(
                filepath, format=format, **table_kwargs
            )

            if name_mapping:
                for name_, name_alt_ in name_mapping.items():
                    self._pdata.rename_column(name_alt_, name_)

        # Validate particle data columns.
        if self._backend == 'nbodykit':
            colnames = self._pdata.columns
        if self._backend == 'astropy':
            colnames = self._pdata.colnames

        for axis_name in ['x', 'y', 'z']:
            if axis_name not in colnames:
                raise MissingValueError(
                    f"Mandatory field {axis_name} is missing."
                )

        if 'nz' not in colnames:
            self._pdata['nz'] = 0.
            warnings.warn(
                "Catalogue 'nz' field is not provided and thus set to zero, "
                "which may raise errors in some computations.",
                category=DefaultValueWarning
            )

        for name_wgt in ['ws', 'wc']:
            if name_wgt not in colnames:
                warnings.warn(
                    f"Catalogue '{name_wgt}' field is not provided, "
                    "so is set to unity.",
                    category=DefaultValueWarning
                )
                self._pdata[name_wgt] = 1.

        # Compute catalogue properties.
        self._calc_bounds(init=True)

        self.ntotal = len(self._pdata)
        self.wtotal = self._compute(
            (self._pdata['ws'] * self._pdata['wc']).sum()
        )
        self.wstotal = self._compute(self._pdata['ws'].sum())

        if self._logger:
            self._logger.info(
                "Catalogue loaded: "
                "ntotal = %d, wtotal = %.3f, wstotal = %.3f (%s).",
                self.ntotal, self.wtotal, self.wstotal, self
            )

        return self

    def __str__(self):
        try:
            return "ParticleCatalogue(source={})".format(self._source)
        except NameError:
            return "ParticleCatalogue(address={})".format(
                object.__str__(self).split()[-1].lstrip().rstrip('>')
            )

    def __len__(self):
        return len(self._pdata)

    def __getitem__(self, key):
        """Return data entry.

        Parameters
        ----------
        key : (list of) str, (list of) int or slice
            Data entry key.

        Returns
        -------
        :class:`numpy.ndarray`
            Data entry.

        """
        return self._compute(self._pdata[key])

    def __setitem__(self, key, val):
        """Set data column.

        Parameters
        ----------
        key : str
            Data column name.
        val : array_like
            Data column array.

        """
        if isinstance(key, str):
            self._pdata[key] = val
        else:
            raise ValueError('Cannot set values for non-string-type key.')

    def compute_los(self):
        """Compute the line of sight to each particle.

        Returns
        -------
        los : (N, 3) :class:`numpy.ndarray`
            Normalised line-of-sight vectors.

        """
        los_norm = np.sqrt(
            self._pdata['x']**2 + self._pdata['y']**2 + self._pdata['z']**2
        )

        los_norm[los_norm == 0.] = 1.

        los = np.transpose([
            self._pdata['x'] / los_norm,
            self._pdata['y'] / los_norm,
            self._pdata['z'] / los_norm
        ])

        return self._compute(los)

    def compute_mean_density(self, volume=None, boxsize=None):
        """Compute the mean density over a volume.

        This sets the 'nz' column to :attr:`ntotal` divided by
        `volume` or cubic (product of) `boxsize`, and is typically used
        for calculating the homogeneous background number density
        in a simulation box.

        Parameters
        ----------
        volume : double, optional
            Volume over which the mean density is calculated.
        boxsize : float or sequence of [float, float, float], optional
            Box size (in each dimension) over which the mean density
            is calculated.  Used only when `volume` is `None`.


        .. attention::

            The invocation of this method resets the particle data
            column ``'nz'``.

        """
        if np.isscalar(boxsize):
            boxsize = [boxsize, boxsize, boxsize]

        volume = volume or np.product(boxsize)

        self._pdata['nz'] = self.ntotal / volume

    def centre(self, boxsize, catalogue_ref=None):
        """Centre a (pair of) catalogue(s) in a box.

        Parameters
        ----------
        boxsize : float or sequence of [float, float, float]
            Box size (in each dimension).
        catalogue_ref : :class:`~triumvirate.catalogue.ParticleCatalogue`, \
                        optional
            Reference catalogue used for box alignment, also to be centred
            in the same box.  If `None` (default), the current catalogue
            itself is used as the reference catalogue.


        .. note::

            The reference catalogue is typically the random-source
            catalogue (if provided).  Particle coordinates in both
            catalogues are shifted by the same displacement vector such
            that the mid-point of particle coordinate extents in the
            reference catalogue is at the centre of the box.

        """
        if np.isscalar(boxsize):
            boxsize = [boxsize, boxsize, boxsize]

        if catalogue_ref is None:
            catalogue_ref_ = self
        else:
            catalogue_ref_ = catalogue_ref

        _axes_overflow = []
        _axes_overflow.extend(self._check_bounds_in_boxsize(boxsize))
        _axes_overflow.extend(catalogue_ref_._check_bounds_in_boxsize(boxsize))
        if _axes_overflow:
            warnings.warn(
                "Box size is smaller than particle coordinate extents "
                "along axis: {}. "
                "Some partcles may lie outside the box after centring."
                .format(set(_axes_overflow))
            )

        origin = np.array([
            np.mean(catalogue_ref_.bounds[axis]) - boxsize[iaxis]/2.
            for iaxis, axis in enumerate(['x', 'y', 'z'])
        ])

        self.offset_coords(origin)
        if catalogue_ref is not None:
            catalogue_ref.offset_coords(origin)

    def pad(self, boxsize, ngrid=None, boxsize_pad=None, ngrid_pad=None,
            catalogue_ref=None):
        """Pad a (pair of) catalogue(s) in a box.

        The particle coordinates are shifted away from the box corner
        at the origin such that in each dimension the minimum particle
        coordinate is given by the amount of padding, which can be set as
        a fraction of the box size or a multiple of the grid cell size.

        Parameters
        ----------
        boxsize : float or sequence of [float, float, float]
            Box size in each dimension.
        ngrid : int or sequence of [int, int, int], optional
            Grid number in each dimension (default is `None`).
            Must be provided if `ngrid_pad` is set.
        boxsize_pad : float or sequence of [float, float, float], optional
            Box size padding factor.  If not `None` (default), then
            `ngrid_pad` must be `None`.
        ngrid_pad : float or sequence of [float, float, float], optional
            Grid padding factor.  If not `None` (default), then
            `boxsize_pad` must be `None`.
        catalogue_ref : :class:`~triumvirate.catalogue.ParticleCatalogue`, \
                        optional
            Reference catalogue used for box alignment, also to be put in
            the same box.  If `None` (default), the current catalogue
            itself is used as the reference catalogue.

        Raises
        ------
        ValueError
            If `boxsize_pad` and `ngrid_pad` are both set to not `None`.


        .. note::

            The reference catalogue is typically the random-source
            catalogue (if provided).  Particle coordinates in both
            catalogues are shifted by the same displacement vector such
            that the minimum particle coordinates in the reference
            catalogue are the amounts of padding specified.

        """
        if boxsize_pad is None and ngrid_pad is None:
            warnings.warn(
                "`boxsize_pad` and `ngrid_pad` are both None. "
                "No padding is applied to the catalogue."
            )
            return
        if boxsize_pad is not None and ngrid_pad is not None:
            raise ValueError(
                "Conflicting padding as `boxsize_pad` and `ngrid_pad` "
                "are both set (not None)."
            )
        if np.isscalar(boxsize):
            boxsize = [boxsize, boxsize, boxsize]

        if catalogue_ref is None:
            catalogue_ref_ = self
        else:
            catalogue_ref_ = catalogue_ref

        _axes_overflow = []
        _axes_overflow.extend(self._check_bounds_in_boxsize(boxsize))
        _axes_overflow.extend(catalogue_ref_._check_bounds_in_boxsize(boxsize))
        if _axes_overflow:
            warnings.warn(
                "Box size is smaller than particle coordinate extents "
                "along axis: {}. "
                "Some partcles may lie outside the box after padding."
                .format(set(_axes_overflow))
            )

        origin = np.array([
            catalogue_ref_.bounds[axis][0] for axis in ['x', 'y', 'z']
        ])

        if boxsize_pad:
            origin -= np.multiply(boxsize_pad, boxsize)
        if ngrid_pad:
            origin -= np.multiply(ngrid_pad, np.divide(boxsize, ngrid))

        self.offset_coords(origin)
        if catalogue_ref is not None:
            catalogue_ref.offset_coords(origin)

        for iaxis, axis in enumerate(['x', 'y', 'z']):
            if (max(self.bounds[axis]) > boxsize[iaxis]
                    or max(catalogue_ref_.bounds[axis]) > boxsize[iaxis]):
                warnings.warn(
                    "`boxsize` is smaller than the particle extents "
                    f"in {axis} axis. Some particles now lie outside "
                    "the box after padding."
                )

    def periodise(self, boxsize):
        """Place particles in a periodic box of given box size.

        Parameters
        ----------
        boxsize : float or sequence of [float, float, float]
            Box size (in each dimension).

        """
        if np.isscalar(boxsize):
            boxsize = [boxsize, boxsize, boxsize]

        _axes_overflow = self._check_bounds_in_boxsize(boxsize)
        if _axes_overflow:
            warnings.warn(
                "Box size is smaller than particle coordinate extents "
                "along axis: {}."
                .format(_axes_overflow)
            )

        for iaxis, axis in enumerate(['x', 'y', 'z']):
            # Also centre.
            self._pdata[axis] = (
                self._pdata[axis] + boxsize[iaxis] / 2. - (
                    self.bounds[axis][1] + self.bounds[axis][0]
                ) / 2.
            ) % boxsize[iaxis]

        self._calc_bounds()

    def offset_coords(self, origin):
        """Offset particle coordinates for a given origin.

        Parameters
        ----------
        origin : float or sequence of [float, float, float]
            Coordinates of the new origin.

        """
        for axis, coord in zip(['x', 'y', 'z'], origin):
            self._pdata[axis] -= coord

        self._calc_bounds()

    def _calc_bounds(self, init=False):
        """Calculate coordinate bounds in each dimension.

        Parameters
        ----------
        init : bool, optional
            If `True` (default is `False`), particle positions are
            original and have not been offset previously.

        """
        self.bounds = {}
        for axis in ['x', 'y', 'z']:
            self.bounds[axis] = (
                self._compute(self._pdata[axis].min()),
                self._compute(self._pdata[axis].max())
            )

        if self._logger:
            if init:
                self._logger.info(
                    "Original extents of particle coordinates: "
                    "{'x': (%.3f, %.3f),"
                    " 'y': (%.3f, %.3f),"
                    " 'z': (%.3f, %.3f)}"
                    " (%s).",
                    *self.bounds['x'], *self.bounds['y'], *self.bounds['z'],
                    self
                )
            else:
                self._logger.info(
                    "Offset extents of particle coordinates: "
                    "{'x': (%.3f, %.3f),"
                    " 'y': (%.3f, %.3f),"
                    " 'z': (%.3f, %.3f)}"
                    " (%s).",
                    *self.bounds['x'], *self.bounds['y'], *self.bounds['z'],
                    self
                )

    def _check_bounds_in_boxsize(self, boxsize):
        """Check if partice coordinate extents are covered by the box size
        in all dimensions, and return the axis name(s) if not.

        Parameters
        ----------
        boxsize : sequence of float
            Box size in all dimensions.

        Returns
        -------
        list of str
            Axis if and where box size is smaller than the particle
            coordinate extent.

        """
        _axes = []
        for iaxis, axis in enumerate(['x', 'y', 'z']):
            if np.abs(np.diff(self.bounds[axis])) > boxsize[iaxis]:
                _axes.append(axis)

        return _axes

    def _convert_to_cpp_catalogue(self, verbose=-1):
        """Convert to a C++-wrapped catalogue.

        Parameters
        ----------
        verbose : int, optional
            Verbosity level to be passed to the backend C++ logger
            (default is -1, i.e. unused).

        Returns
        -------
        :class:`~triumvirate._particles._ParticleCatalogue`
            C++-wrapped catalogue.

        """
        x = self._compute(self._pdata['x'])
        y = self._compute(self._pdata['y'])
        z = self._compute(self._pdata['z'])
        nz = self._compute(self._pdata['nz'])
        ws = self._compute(self._pdata['ws'])
        wc = self._compute(self._pdata['wc'])

        return _ParticleCatalogue(x, y, z, nz, ws, wc, verbose=verbose)

    def _compute(self, quant):
        """Return a quantity in standard form (i.e. apply
        :meth:`dask.array.Array.compute` to any Dask array).

        Parameters
        ----------
        quant
            Quantity.

        Returns
        -------
        The same quantity in a non-Dask-array form.

        """
        try:
            return quant.compute()
        except AttributeError:
            return quant

    def write_attrs_as_header(self, catalogue_ref=None):
        """Write catalogue attributes as a header.

        Parameters
        ----------
        catalogue_ref : :class:`~triumvirate.catalogue.ParticleCatalogue`, \
                        optional
            Reference catalogue (default is `None`) whose attributes are
            also written out.  This is typically the
            random-source catalogue.

        Returns
        -------
        text_header : str
            Catalogue attributes as a header string.

        """
        if catalogue_ref is None:
            text_lines = [
                "Catalogue source: {}"
                    .format(self._source),  # noqa: E131
                "Catalogue size: "
                "ntotal = {:d}, wtotal = {:.3f}, wstotal = {:.3f}"
                    .format(self.ntotal, self.wtotal, self.wstotal),
                "Catalogue particle extents: "
                "([{:.3f}, {:.3f}], [{:.3f}, {:.3f}], [{:.3f}, {:.3f}])"
                    .format(
                        *self.bounds['x'], *self.bounds['y'], *self.bounds['z']
                    ),
            ]
        else:
            text_lines = [
                "Data catalogue source: {}"
                    .format(self._source),  # noqa: E131
                "Data catalogue size: "
                "ntotal = {:d}, wtotal = {:.3f}, wstotal = {:.3f}"
                    .format(self.ntotal, self.wtotal, self.wstotal),
                "Data-source particle extents: "
                "([{:.3f}, {:.3f}], [{:.3f}, {:.3f}], [{:.3f}, {:.3f}])"
                    .format(
                        *self.bounds['x'], *self.bounds['y'], *self.bounds['z']
                    ),
                "Random catalogue source: {}"
                    .format(catalogue_ref._source),  # noqa: E131
                "Random catalogue size: "
                "ntotal = {:d}, wtotal = {:.3f}, wstotal = {:.3f}"
                    .format(  # noqa: E131
                        catalogue_ref.ntotal,
                        catalogue_ref.wtotal,
                        catalogue_ref.wstotal,
                    ),
                "Random-source particle extents: "
                "([{:.3f}, {:.3f}], [{:.3f}, {:.3f}], [{:.3f}, {:.3f}])"
                    .format(  # noqa: E131
                        *catalogue_ref.bounds['x'],
                        *catalogue_ref.bounds['y'],
                        *catalogue_ref.bounds['z']
                    ),
            ]

        text_header = "\n".join(text_lines)

        return text_header



"""Test :mod:`~triumvirate.logger`.

"""
import re

import pytest

from triumvirate.logger import setup_logger

# Define the RegEx expression for logging format.
_LOG_FORMAT = (
    r"\[\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\s\(\+\d+:\d{2}:\d{2}\)\s\w{4}\]"
    r"\s.*\(in C\+\+\)"
)


@pytest.fixture
def default_logger():
    return setup_logger()


@pytest.mark.no_capture
def test_setup_logger_formatter(default_logger, capfd):
    # Test logging formatter with C++ code state indication.
    default_logger.info("Test log entry.", cpp_state=True)
    assert re.match(_LOG_FORMAT, capfd.readouterr().out) is not None, \
        "Misformatted logging."


def test_setup_logger_adapter(default_logger, caplog):
    # Test logger C++ code state indication only.
    default_logger.info("Test log entry with C++ indication.", cpp_state=True)
    assert re.match(r'.*\(in C\+\+\)', caplog.text) is not None, \
        "No C++ indication shown by logger."



"""
Data Objects (:mod:`~triumvirate.dataobjs`)
==========================================================================

Define data objects.

.. autosummary::
    Binning

"""
import numpy as np

from .dataobjs cimport CppBinning


cdef class Binning:
    """Binning.

    Parameters
    ----------
    space : {'config', 'fourier'}
        Coordinate space.
    scheme : {'lin', 'log', 'linpad', 'logpad'}
        Binning scheme.
    bin_min, bin_max : float, optional
        Minimum and maximum of the bin range (defaults are `None`).
    num_bins : int, optional
        Number of bins (default is `None`).

    Attributes
    ----------
    space : {'config', 'fourier'}
        Coordinate space.
    scheme : {'lin', 'log', 'linpad', 'logpad'}
        Binning scheme.
    bin_min, bin_max : float or None
        Minimum and maximum of the bin range.
    num_bins : int or None
        Number of bins.
    bin_edges : list of float
        Bin edges of length (:attr:`num_bins` + 1).
    bin_centres : list of float
        Bin centres of length :attr:`num_bins`.
    bin_widths : list of float
        Bin widths of length :attr:`num_bins`.


    .. note::

        The bin setting method supports linear ('lin') and
        log-linear/exponential ('log') binning, with possible linear
        padding from zero for the first 5 bins ('linpad' or 'logpad').
        The padding is either 1.e-3 ('fourier' space) and 10. ('config'
        space), or determined by the mesh grid resolution if set using
        mesh grid sizes.

    """

    def __cinit__(self, space, scheme,
                  bin_min=None, bin_max=None, num_bins=None):

        self.thisptr = new CppBinning(
            space.encode('utf-8'), scheme.encode('utf-8')
        )

        self.scheme = scheme
        self.space = space

        _settable = True
        if bin_min is not None:
            self.bin_min = bin_min
        else:
            _settable = False
        if bin_max is not None:
            self.bin_max = bin_max
        else:
            _settable = False
        if num_bins is not None:
            self.num_bins = num_bins
        else:
            _settable = False

        if _settable:
            self.set_bins(self.bin_min, self.bin_max, self.num_bins)

    @classmethod
    def from_parameter_set(cls, paramset):
        """Create binning scheme from a parameter set.

        This sets :attr:`scheme` and :attr:`space`.

        Parameters
        ----------
        paramset : :class:`~triumvirate.parameters.ParameterSet`
            Parameter set.

        Raises
        ------
        ValueError
            When the 'space' parameter is unset or unrecognised in
            `paramset`.


        .. attention::

            If `paramset` is initialised without an accepted
            'statistic_type' parameter value, the derived 'space'
            parameter is unset and must be specified in
            {'fourier', 'config'} before passing `paramset` to this
            method. See also
            :class:`~triumvirate.parameters.ParameterSet`.

        """
        if paramset['space'] not in ['fourier', 'config']:
            raise ValueError(
                "'space' parameter is unset or unrecognised in `paramset`."
            )

        self = cls(
            paramset['space'], paramset['binning'],
            bin_min=paramset['range'][0],
            bin_max=paramset['range'][-1],
            num_bins=paramset['num_bins']
        )

        return self

    def set_bins(self, bin_min, bin_max, num_bins):
        """Set bin quantities including recalculated bin edges, centres
        and widths.

        Parameters
        ----------
        bin_min, bin_max : float
            Minimum and maximum of the bin range.
        num_bins : int
            Number of bins.

        """
        self.bin_min, self.bin_max = bin_min, bin_max
        self.num_bins = num_bins

        self.thisptr.set_bins(self.bin_min, self.bin_max, self.num_bins)

        self.bin_edges = self.thisptr.bin_edges
        self.bin_centres = self.thisptr.bin_centres
        self.bin_widths = self.thisptr.bin_widths

    def set_grid_based_bins(self, boxsize, ngrid):
        """Set linear binning based on a mesh grid.

        The binning scheme is overriden to 'lin'.  The bin width is
        given by the grid resolution in configuration space or the
        fundamental wavenumber in Fourier space.  The bin minimum is zero
        and the bin maximum is half the boxsize in configuration space or
        the Nyquist wavenumber in Fourier space.

        Parameters
        ----------
        boxsize : float or sequence of [float, float, float]
            Mesh box size (maximum).
        ngrid : int or sequence of [int, int, int]
            Mesh grid number (minimum).

        """
        boxsize = np.asarray(boxsize).max()
        ngrid = np.asarray(ngrid).min()

        self.thisptr.set_bins(boxsize, ngrid)

        self.bin_min, self.bin_max = self.thisptr.bin_min, self.thisptr.bin_max
        self.num_bins = self.thisptr.num_bins
        self.bin_edges = self.thisptr.bin_edges
        self.bin_centres = self.thisptr.bin_centres
        self.bin_widths = self.thisptr.bin_widths
        self.scheme = self.thisptr.scheme

    def set_custom_bins(self, bin_edges):
        """Set custom bins using bin edges.

        :attr:`bin_centres` and :attr:`bin_widths` are also automatically
        set for internal consistency. :attr:`scheme` is set to 'custom'.

        Parameters
        ----------
        bin_edges : 1-d array of float
            Bin edges of length (:attr:`num_bins` + 1).

        """
        self.bin_min, self.bin_max = bin_edges[0], bin_edges[-1]
        self.num_bins = len(bin_edges) - 1

        self.bin_edges = np.array(bin_edges)
        self.bin_centres = np.add(bin_edges[:-1], bin_edges[1:]) / 2.
        self.bin_widths = np.subtract(bin_edges[1:], bin_edges[:-1])

        self.scheme = 'custom'

        self.thisptr.bin_min = self.bin_min
        self.thisptr.bin_max = self.bin_max
        self.thisptr.num_bins = self.num_bins
        self.thisptr.bin_edges = self.bin_edges
        self.thisptr.bin_centres = self.bin_centres
        self.thisptr.bin_widths = self.bin_widths
        self.thisptr.scheme = self.scheme.encode('utf-8')



"""
Three-Point Correlator Algorithms (:mod:`~triumvirate._threept`)
==========================================================================

Declare and interface with three-point correlator algorithms.

"""
from cython.operator cimport dereference as deref
from libc.stdlib cimport free, malloc
from libcpp cimport bool as bool_t

import numpy as np
cimport numpy as np

from triumvirate._particles cimport CppParticleCatalogue, _ParticleCatalogue
from triumvirate.dataobjs cimport (
    Binning, CppBinning,
    LineOfSight,
    BispecMeasurements, ThreePCFMeasurements, ThreePCFWindowMeasurements
)
from triumvirate.parameters cimport CppParameterSet, ParameterSet


cdef extern from "include/threept.hpp":
    # --------------------------------------------------------------------
    # Normalisation
    # --------------------------------------------------------------------

    double calc_bispec_normalisation_from_particles_cpp \
        "trv::calc_bispec_normalisation_from_particles" (
            CppParticleCatalogue& particles,
            double alpha
        ) except +

    double calc_bispec_normalisation_from_mesh_cpp \
        "trv::calc_bispec_normalisation_from_mesh" (
            CppParticleCatalogue& particles,
            CppParameterSet& params,
            double alpha
        )


    # --------------------------------------------------------------------
    # Full correlator
    # --------------------------------------------------------------------

    BispecMeasurements compute_bispec_cpp "trv::compute_bispec" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& kbinning,
        double norm_factor
    )

    ThreePCFMeasurements compute_3pcf_cpp "trv::compute_3pcf" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& rbinning,
        double norm_factor
    )

    BispecMeasurements compute_bispec_in_gpp_box_cpp \
        "trv::compute_bispec_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& kbinning,
            double norm_factor
        )

    ThreePCFMeasurements compute_3pcf_in_gpp_box_cpp \
        "trv::compute_3pcf_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& rbinning,
            double norm_factor
        )

    ThreePCFWindowMeasurements compute_3pcf_window_cpp \
        "trv::compute_3pcf_window" (
            CppParticleCatalogue& catalogue_rand,
            LineOfSight* los_rand,
            CppParameterSet& params,
            CppBinning& rbinning,
            double alpha,
            double norm_factor,
            bool_t wide_angle
        )

    # BispecMeasurements compute_bispec_for_los_choice_cpp \
    #     "trv::compute_bispec_for_los_choice" (
    #         CppParticleCatalogue& catalogue_data,
    #         CppParticleCatalogue& catalogue_rand,
    #         LineOfSight* los_data,
    #         LineOfSight* los_rand,
    #         int los_choice,
    #         CppParameterSet& params,
    #         CppBinning& kbinning,
    #         double norm_factor
    #     )


def _calc_bispec_normalisation_from_particles(
        _ParticleCatalogue particles not None, double alpha
    ):
    return calc_bispec_normalisation_from_particles_cpp(
        deref(particles.thisptr), alpha
    )


def _calc_bispec_normalisation_from_mesh(
        _ParticleCatalogue particles not None,
        ParameterSet params not None,
        double alpha
    ):
    return calc_bispec_normalisation_from_mesh_cpp(
        deref(particles.thisptr), deref(params.thisptr), alpha
    )


def _compute_bispec(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef BispecMeasurements results
    results = compute_bispec_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'k1_bin': np.array(results.k1_bin),
        'k2_bin': np.array(results.k2_bin),
        'k1_eff': np.array(results.k1_eff),
        'k2_eff': np.array(results.k2_eff),
        'nmodes': np.array(results.nmodes),
        'bk_raw': np.array(results.bk_raw),
        'bk_shot': np.array(results.bk_shot),
    }


def _compute_3pcf(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef ThreePCFMeasurements results
    results = compute_3pcf_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'r1_bin': np.array(results.r1_bin),
        'r2_bin': np.array(results.r2_bin),
        'r1_eff': np.array(results.r1_eff),
        'r2_eff': np.array(results.r2_eff),
        'npairs': np.array(results.npairs),
        'zeta_raw': np.array(results.zeta_raw),
        'zeta_shot': np.array(results.zeta_shot),
    }


def _compute_bispec_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    cdef BispecMeasurements results
    results = compute_bispec_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    return {
        'k1_bin': np.array(results.k1_bin),
        'k2_bin': np.array(results.k2_bin),
        'k1_eff': np.array(results.k1_eff),
        'k2_eff': np.array(results.k2_eff),
        'nmodes': np.array(results.nmodes),
        'bk_raw': np.array(results.bk_raw),
        'bk_shot': np.array(results.bk_shot),
    }


def _compute_3pcf_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    cdef ThreePCFMeasurements results
    results = compute_3pcf_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    return {
        'r1_bin': np.array(results.r1_bin),
        'r2_bin': np.array(results.r2_bin),
        'r1_eff': np.array(results.r1_eff),
        'r2_eff': np.array(results.r2_eff),
        'npairs': np.array(results.npairs),
        'zeta_raw': np.array(results.zeta_raw),
        'zeta_shot': np.array(results.zeta_shot),
    }


def _compute_3pcf_window(
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double alpha,
        double norm_factor,
        bool_t wide_angle
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef ThreePCFWindowMeasurements results
    results = compute_3pcf_window_cpp(
        deref(catalogue_rand.thisptr), los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        alpha, norm_factor,
        wide_angle
    )

    free(los_rand_cpp)

    return {
        'r1_bin': np.array(results.r1_bin),
        'r2_bin': np.array(results.r2_bin),
        'r1_eff': np.array(results.r1_eff),
        'r2_eff': np.array(results.r2_eff),
        'npairs': np.array(results.npairs),
        'zeta_raw': np.array(results.zeta_raw),
        'zeta_shot': np.array(results.zeta_shot),
    }


# def _compute_bispec_for_los_choice(
#         _ParticleCatalogue catalogue_data not None,
#         _ParticleCatalogue catalogue_rand not None,
#         np.ndarray[double, ndim=2, mode='c'] los_data not None,
#         np.ndarray[double, ndim=2, mode='c'] los_rand not None,
#         int los_choice,
#         ParameterSet params not None,
#         Binning kbinning not None,
#         double norm_factor
#     ):
#     # Parse lines of sight per particle.
#     cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
#         len(los_data) * sizeof(LineOfSight)
#     )
#     for pid, (los_x, los_y, los_z) in enumerate(los_data):
#         los_data_cpp[pid].pos[0] = los_x
#         los_data_cpp[pid].pos[1] = los_y
#         los_data_cpp[pid].pos[2] = los_z

#     cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
#         len(los_rand) * sizeof(LineOfSight)
#     )
#     for pid, (los_x, los_y, los_z) in enumerate(los_rand):
#         los_rand_cpp[pid].pos[0] = los_x
#         los_rand_cpp[pid].pos[1] = los_y
#         los_rand_cpp[pid].pos[2] = los_z

#     # Run algorithm.
#     cdef BispecMeasurements results
#     results = compute_bispec_for_los_choice_cpp(
#         deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
#         los_data_cpp, los_rand_cpp, los_choice,
#         deref(params.thisptr), deref(kbinning.thisptr),
#         norm_factor
#     )

#     free(los_data_cpp); free(los_rand_cpp)

#     return {
#         'k1_bin': np.array(results.k1_bin),
#         'k2_bin': np.array(results.k2_bin),
#         'k1_eff': np.array(results.k1_eff),
#         'k2_eff': np.array(results.k2_eff),
#         'nmodes': np.array(results.nmodes),
#         'bk_raw': np.array(results.bk_raw),
#         'bk_shot': np.array(results.bk_shot),
#     }



"""
Program Logger (:mod:`~triumvirate.logger`)
==========================================================================

Configure the program logger.

.. autosummary::
    setup_logger

"""
import logging
import sys
import time
import warnings


class _ElapsedLogFormatter(logging.Formatter):
    """Elapsed-time logging formatter.

    """

    _start_time = time.time()

    def format(self, record):
        """Add elapsed time in hours, minutes and seconds to the log
        record.

        Parameters
        ----------
        record : :class:`Logging.LogRecord`
            See :class:`logging.LogRecord`.

        Returns
        -------
        str
            Modified log record with elapsed time.

        """
        elapsed_time = record.created - self._start_time
        h, remainder_time = divmod(elapsed_time, 3600)
        m, s = divmod(remainder_time, 60)

        record.elapsed = \
            "(+{:02d}:{:02d}:{:02d})".format(int(h), int(m), int(s))

        return logging.Formatter.format(self, record)


class _CppLogAdapter(logging.LoggerAdapter):
    """C++ logging adapter.

    """

    def process(self, msg, kwargs):
        """Adapt logger message.

        Parameters
        ----------
        msg : str
            See :class:`logging.LoggerAdapter`.
        kwargs : dict
            See :class:`logging.LoggerAdapter`.

        Returns
        -------
        str
            Adapted log message with C++ runtime indication.

        """
        # Extract passed state variable or resort to default from `extra`.
        cpp_state = kwargs.pop('cpp_state', self.extra['cpp_state'])

        if isinstance(cpp_state, str):
            if cpp_state.lower() == 'start':
                return "%s (entering C++)" % msg, kwargs
            if cpp_state.lower() == 'end':
                return "%s (exited C++)" % msg, kwargs

        if cpp_state:  # merely indicating in CPP state, no extra text
            return "%s (in C++)" % msg, kwargs
        return "%s" % msg, kwargs


# Modify `warnings.formatwarning`.
def _format_warning(message, category, filename, lineno, line=None):
    """Modify formatter for warning messages in logged record.

    See :func:`warnings.formatwarning` for implementation details.

    """
    msg = warnings.WarningMessage(
        message, category, filename, lineno, None, line
    )

    msg_txt = "{} ({}:{}:{}{{}})".format(
        str(msg.message).replace('{', '{{').replace('}', '}}'),
        msg.filename, msg.lineno, msg.category.__name__
    )

    if msg.line is None:
        try:
            import linecache
            linesource = linecache.getline(msg.filename, msg.lineno)
        except Exception:
            linesource = None
    else:
        linesource = msg.line

    if linesource:
        linesource = linesource.strip()
        msg_txt = msg_txt.format(" --> \"" + linesource + "\"")
    else:
        msg_txt = msg_txt.format("")

    return msg_txt


def setup_logger(log_level=logging.INFO):
    """Set up and return a customised logger with elapsed time,
    C++ runtime indication and formatted warning messages.

    Parameters
    ----------
    log_level : int, optional
        Logging devel (default is `logging.INFO`).

    Returns
    -------
    logger : :class:`logging.LoggerAdapter`
        Customised logger.

    """
    # Set formatter.
    formatter = _ElapsedLogFormatter(
        fmt='[%(asctime)s %(elapsed)s %(levelname).4s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    # Set handler.
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    # Instantiate logger.
    logger_ = logging.getLogger()

    if logger_.hasHandlers():
        logger_.handlers.clear()
    logger_.addHandler(handler)

    logger_.setLevel(log_level)

    # Adapt logger for C++ code indication.
    logger = _CppLogAdapter(logger_, {'cpp_state': False})

    # Adapt logger to capture warnings.
    warnings.formatwarning = _format_warning
    logging.captureWarnings(True)

    return logger



"""
FFTLog algorithm (:mod:`~triumvirate._fftlog`)
==========================================================================

Implement the FFTLog algorithm for Hankel-related transforms.

"""
import numpy as np
cimport numpy as np


cdef extern from "include/fftlog.hpp":
    int sj_transform_cpp "trv::maths::sj_transform" (
        int ell, int m, int N, double* r, double* a, double* k, double* b
    )

    int sj_transform_symm_biased_cpp "trv::maths::sj_transform_symm_biased" (
        int ell, int i, int N, double* r, double* a, double* k, double* b
    )

    int transform_powspec_to_corrfunc_multipole_cpp \
        "trv::transform_powspec_to_corrfunc_multipole" (
            int ell, int N, double* k, double* pk, double* r, double* xi
        )

    int transform_corrfunc_to_powspec_multipole_cpp \
        "trv::transform_corrfunc_to_powspec_multipole" (
            int ell, int N, double* r, double* xi, double* k, double* pk
        )


def sj_transform(
        int ell, int m, int N,
        np.ndarray[double, ndim=1, mode='c'] x not None,
        np.ndarray[double, ndim=1, mode='c'] fx not None
    ):
    cdef np.ndarray[double, ndim=1, mode='c'] y = np.zeros(N)
    cdef np.ndarray[double, ndim=1, mode='c'] gy = np.zeros(N)

    sj_transform_cpp(ell, m, N, &x[0], &fx[0], &y[0], &gy[0])

    return y, gy


def sj_transform_symm_biased(
        int ell, int i, int N,
        np.ndarray[double, ndim=1, mode='c'] x not None,
        np.ndarray[double, ndim=1, mode='c'] fx not None
    ):
    cdef np.ndarray[double, ndim=1, mode='c'] y = np.zeros(N)
    cdef np.ndarray[double, ndim=1, mode='c'] gy = np.zeros(N)

    sj_transform_symm_biased_cpp(ell, i, N, &x[0], &fx[0], &y[0], &gy[0])

    return y, gy


def trans_powspec_to_corrfunc_multipole(
        int ell, int N,
        np.ndarray[double, ndim=1, mode='c'] k not None,
        np.ndarray[double, ndim=1, mode='c'] pk not None
    ):
    cdef np.ndarray[double, ndim=1, mode='c'] r = np.zeros(N)
    cdef np.ndarray[double, ndim=1, mode='c'] xi = np.zeros(N)

    transform_powspec_to_corrfunc_multipole_cpp(
        ell, N, &k[0], &pk[0], &r[0], &xi[0]
    )

    return r, xi


def trans_corrfunc_to_powspec_multipole(
        int ell, int N,
        np.ndarray[double, ndim=1, mode='c'] r not None,
        np.ndarray[double, ndim=1, mode='c'] xi not None
    ):
    cdef np.ndarray[double, ndim=1, mode='c'] k = np.zeros(N)
    cdef np.ndarray[double, ndim=1, mode='c'] pk = np.zeros(N)

    transform_corrfunc_to_powspec_multipole_cpp(
        ell, N, &r[0], &xi[0], &k[0], &pk[0]
    )

    return k, pk



"""Test :mod:`~triumvirate.catalogue`.

"""
import warnings

import numpy as np
import pytest

from triumvirate.catalogue import (
    DefaultValueWarning,
    ParticleCatalogue,
)

CONST_LEN = 1
CONST_COORDS = np.ones(CONST_LEN)
CONST_NZ = 1.e-3

SAMPLE_LEN = 1
SAMPLE_COORDS = {
    axis: [coord,] * SAMPLE_LEN  # noqa: E231
    for axis, coord in zip(['x', 'y', 'z'], [-1., 0., 1.])
}
SAMPLE_NZ = 1.e-3
SAMPLE_WSYS = 1.1
SAMPLE_WFKP = 0.9


@pytest.fixture
def minimal_catalogue():
    x = y = z = CONST_COORDS
    nz = None
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message=".*'nz' field is None.*")
        return ParticleCatalogue(x, y, z, nz=nz)


@pytest.fixture
def minimal_catalogue_paired():
    x = y = CONST_COORDS
    z = CONST_COORDS - 0.5
    nz = CONST_NZ
    return ParticleCatalogue(x, y, z, nz=nz)


@pytest.mark.parametrize(
    "x, y, z, nz, ws, wc",
    [
        (np.arange(5), -np.arange(5), np.arange(5) / 2., 1.e-3, 1.1, 0.9),
        (np.arange(5), -np.arange(5), np.arange(5) / 2., None, None, None),
    ]
)
def test_ParticleCatalogue___init__(x, y, z, nz, ws, wc):

    weight_kwargs = {}
    if ws is not None:
        weight_kwargs.update(ws=ws)
    if wc is not None:
        weight_kwargs.update(wc=wc)

    if nz is None:
        # Check warning is raised for missing the 'nz' field.
        with pytest.warns(DefaultValueWarning):
            ParticleCatalogue(x, y, z, nz=nz, **weight_kwargs)
    else:
        ParticleCatalogue(x, y, z, nz=nz, **weight_kwargs)


@pytest.mark.parametrize(
    "filename, reader, format, names, name_mapping",
    [
        (
            "sample_catalogue.txt",
            'astropy',
            'ascii.no_header',
            ['x', 'y', 'z', 'nz', 'ws', 'wc'],
            None,
        ),
        (
            "sample_catalogue.txt",
            'nbodykit',
            'text',
            ['x', 'y', 'z', 'nz', 'ws', 'wc'],
            None,
        ),
        (
            "sample_catalogue.fits",
            'astropy',
            'fits',
            None,
            {
                'x': 'X', 'y': 'Y', 'z': 'Z',
                'nz': 'NZ', 'ws': 'WEIGHT_SYS', 'wc': 'WEIGHT_FKP'
            },
        ),
        (
            "sample_catalogue.dat",
            'nbodykit',
            'binary',
            [
                ('X', np.float64), ('Y', np.float64), ('Z', np.float64),
                ('NZ', np.float64),
                ('WEIGHT_SYS', np.float64), ('WEIGHT_FKP', np.float64),
            ],
            {
                'x': 'X', 'y': 'Y', 'z': 'Z',
                'nz': 'NZ', 'ws': 'WEIGHT_SYS', 'wc': 'WEIGHT_FKP'
            },
        ),
        (
            "sample_catalogue.h5",
            'nbodykit',
            'hdf5',
            None,
            {
                'x': 'X', 'y': 'Y', 'z': 'Z',
                'nz': 'NZ', 'ws': 'WEIGHT_SYS', 'wc': 'WEIGHT_FKP'
            },
        ),
    ]
)
def test_ParticleCatalogue_read_from_file(filename, reader, format,
                                          names, name_mapping, test_ctlg_dir):

    if reader == 'nbodykit':
        import importlib
        try:
            importlib.import_module(reader)
        except ImportError:
            return

    catalogue = ParticleCatalogue.read_from_file(
        test_ctlg_dir/filename,
        reader=reader, names=names, format=format, name_mapping=name_mapping
    )

    for axis in ['x', 'y', 'z']:
        assert catalogue[axis] == SAMPLE_COORDS[axis], \
            "Loaded catalogue particle coordinates do not match."
    assert catalogue['nz'] == np.ones(CONST_LEN) * SAMPLE_NZ, \
        "Loaded catalogue particle number density column does not match."
    assert catalogue['ws'] == np.ones(CONST_LEN) * SAMPLE_WSYS, \
        "Loaded catalogue particle sample weight column does not match."
    assert catalogue['wc'] == np.ones(CONST_LEN) * SAMPLE_WFKP, \
        "Loaded catalogue particle clustering weight column does not match."


def test_ParticleCatalogue___str__(minimal_catalogue):
    assert 'extdata' in str(minimal_catalogue), \
        "Catalogue string representation has incorrect source."


def test_ParticleCatalogue___len__(minimal_catalogue):
    assert len(minimal_catalogue) == CONST_LEN, \
        "Catalogue length is incorrect."


def test_ParticleCatalogue___getitem__(minimal_catalogue):
    assert np.allclose(minimal_catalogue['x'], CONST_COORDS), \
        "Catalogue column not returned correctly."


def test_ParticleCatalogue_compute_los(minimal_catalogue):
    _los = np.tile(CONST_COORDS, (CONST_LEN, 3)) / np.linalg.norm(
        np.tile(CONST_COORDS, (CONST_LEN, 3)), axis=-1
    )
    assert np.allclose(minimal_catalogue.compute_los(), _los), \
        "Catalogue lines of sight not computed correctly."


@pytest.mark.parametrize(
    "volume, boxsize, nz",
    [
        (1.e3, None, CONST_NZ),
        (None, 10., CONST_NZ),
        (None, [10.,]*3, CONST_NZ),  # noqa: E231
    ]
)
def test_ParticleCatalogue_compute_mean_density(
    volume, boxsize, nz, minimal_catalogue
):
    minimal_catalogue.compute_mean_density(volume=volume, boxsize=boxsize)
    assert all(np.isclose(minimal_catalogue['nz'], nz)), \
        "Catalogue in-box mean density not computed correctly."


@pytest.mark.parametrize(
    "boxsize, ref_catalogue, extents",
    [
        (
            1000.,
            None,
            dict(zip(['x', 'y', 'z'], [[500.,]*2]*3)),  # noqa: E231
        ),
        (
            1000.,
            'minimal_catalogue_paired',
            dict(
                zip(
                    ['x', 'y', 'z'], [[500.,]*2]*2 + [[500.5,]*2]  # noqa: E231
                )
            ),
        ),
    ]
)
def test_ParticleCatalogue_centre(boxsize, ref_catalogue, extents,
                                  minimal_catalogue, request):

    if ref_catalogue is not None:
        ref_catalogue = request.getfixturevalue(ref_catalogue)

    minimal_catalogue.centre(boxsize, catalogue_ref=ref_catalogue)
    for axis in ['x', 'y', 'z']:
        assert np.allclose(minimal_catalogue.bounds[axis], extents[axis]), \
            "Catalogue particles are not centred correctly in the box."


@pytest.mark.parametrize(
    "boxsize, ngrid, boxsize_pad, ngrid_pad, ref_catalogue, extents",
    [
        (
            1000.,
            None,
            0.05,
            None,
            None,
            dict(zip(['x', 'y', 'z'], [[50.,]*2]*3)),  # noqa: E231
        ),
        (
            1000.,
            None,
            0.05,
            None,
            'minimal_catalogue_paired',
            dict(
                zip(
                    ['x', 'y', 'z'], [[50.,]*2]*2 + [[50.5,]*2]  # noqa: E231
                )
            ),
        ),
        (
            1000.,
            10,
            None,
            0.5,
            None,
            dict(zip(['x', 'y', 'z'], [[50.,]*2]*3)),  # noqa: E231
        ),
    ]
)
def test_ParticleCatalogue_pad(boxsize, ngrid, boxsize_pad, ngrid_pad,
                               ref_catalogue, extents,
                               minimal_catalogue, request):

    if ref_catalogue is not None:
        ref_catalogue = request.getfixturevalue(ref_catalogue)

    minimal_catalogue.pad(
        boxsize, ngrid=ngrid, boxsize_pad=boxsize_pad, ngrid_pad=ngrid_pad,
        catalogue_ref=ref_catalogue
    )

    for axis in ['x', 'y', 'z']:
        assert np.allclose(minimal_catalogue.bounds[axis], extents[axis]), \
            "Catalogue particles are not padded correctly in the box."


@pytest.mark.parametrize(
    "boxsize, extents",
    [
        (
            0.3,
            dict(
                zip(
                    ['x', 'y', 'z'],
                    [[0.3/2.,]*2]*3  # noqa: E231
                )
            ),
        ),
    ]
)
def test_ParticleCatalogue_periodise(boxsize, extents, minimal_catalogue):
    minimal_catalogue.periodise(boxsize)
    for axis in ['x', 'y', 'z']:
        assert np.allclose(minimal_catalogue.bounds[axis], extents[axis]), \
            "Catalogue particle coordinates do not satisfy periodicity."


@pytest.mark.parametrize(
    "origin, extents",
    [
        (
            [1., 1., 1.],
            dict(zip(['x', 'y', 'z'], [[0.,]*2]*3)),  # noqa: E231
        ),
    ]
)
def test_ParticleCatalogue_offset_coords(origin, extents, minimal_catalogue):
    minimal_catalogue.offset_coords(origin)
    for axis in ['x', 'y', 'z']:
        assert np.allclose(minimal_catalogue.bounds[axis], extents[axis]), \
            "Catalogue particle coordinates are not offset correctly."


@pytest.mark.parametrize(
    "ref_catalogue",
    [
        'minimal_catalogue_paired',
        None,
    ]
)
def test_ParticleCatalogue_write_attrs_as_header(ref_catalogue,
                                                 minimal_catalogue, request):

    if ref_catalogue is not None:
        ref_catalogue = request.getfixturevalue(ref_catalogue)

    header = minimal_catalogue.write_attrs_as_header(
        catalogue_ref=ref_catalogue
    )

    if ref_catalogue is None:
        assert (
            "Catalogue source: {}".format(minimal_catalogue._source)
        ) in header
        assert (
            "Catalogue size: ntotal = {:d}, wtotal = {:.3f}, wstotal = {:.3f}"
            .format(
                minimal_catalogue.ntotal,
                minimal_catalogue.wtotal,
                minimal_catalogue.wstotal,
            )
        ) in header
        assert (
            "Catalogue particle extents: "
            "([{:.3f}, {:.3f}], [{:.3f}, {:.3f}], [{:.3f}, {:.3f}])"
            .format(
                *minimal_catalogue.bounds['x'],
                *minimal_catalogue.bounds['y'],
                *minimal_catalogue.bounds['z'],
            )
        ) in header
    else:
        for _source_type, _catalogue in zip(
            ['Data', 'Random'], [minimal_catalogue, ref_catalogue]
        ):
            assert (
                "{} catalogue source: {}"
                .format(_source_type, _catalogue._source)
            ) in header
            assert (
                "{} catalogue size: "
                "ntotal = {:d}, wtotal = {:.3f}, wstotal = {:.3f}"
                .format(
                    _source_type,
                    _catalogue.ntotal, _catalogue.wtotal, _catalogue.wstotal,
                )
            ) in header
            assert (
                "{}-source particle extents: "
                "([{:.3f}, {:.3f}], [{:.3f}, {:.3f}], [{:.3f}, {:.3f}])"
                .format(
                    _source_type,
                    *_catalogue.bounds['x'],
                    *_catalogue.bounds['y'],
                    *_catalogue.bounds['z'],
                )
            ) in header



"""
Two-Point Clustering Statistics (:mod:`~triumvirate.twopt`)
==========================================================================

Measure two-point clustering statistics from catalogues.

Local plane-parallel estimators
-------------------------------

.. autosummary::
    compute_powspec
    compute_corrfunc

Global plane-parallel estimators
--------------------------------

.. autosummary::
    compute_powspec_in_gpp_box
    compute_corrfunc_in_gpp_box

Window function estimator
-------------------------

.. autosummary::
    compute_corrfunc_window

"""
import warnings
from pathlib import Path

import numpy as np

from triumvirate._twopt import (
    _calc_powspec_normalisation_from_mesh,
    _calc_powspec_normalisation_from_meshes,
    _calc_powspec_normalisation_from_particles,
    _compute_corrfunc,
    _compute_corrfunc_in_gpp_box,
    _compute_corrfunc_window,
    _compute_powspec,
    _compute_powspec_in_gpp_box,
)
from triumvirate.dataobjs import Binning
from triumvirate.parameters import (
    _modify_measurement_parameters,
    _modify_sampling_parameters,
    fetch_paramset_template,
    ParameterSet,
)


# Use default parameters for mixed-mesh normalisation in `pypower`.
PADDING = .10
CELLSIZE = 10.
ASSIGNMENT = 'cic'


def _amalgamate_parameters(paramset=None, params_sampling=None,
                           degree=None, binning=None, **type_kwargs):
    """Amalgamate a parameter set with overriding sampling parameters
    and the measured multipole degree and coordinate binning.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), the other parameters
        must be provided.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~.triumvirate.dataobjs.Binning`, optional
        Binning (default is `None`).
    **type_kwargs
        `catalogue_type` and `statistic_type` parameters to be filled in.

    Returns
    -------
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Amalgamated full parameter set.

    Raises
    ------
    ValueError
        When `paramset` is `None` while `params_sampling`, `degree` or
        `binning` is also `None`.

    """
    if paramset is None and params_sampling is None:
        raise ValueError(
            "Either `paramset` or `sampling_params` must be provided."
        )
    if paramset is None and degree is None:
        raise ValueError("Either `paramset` or `degree` must be provided.")
    if paramset is None and binning is None:
        raise ValueError("Either `paramset` or `binning` must be provided.")

    if paramset is None:
        paramset, defaults = fetch_paramset_template('dict', ret_defaults=True)
    else:
        # paramset = paramset
        defaults = None

    if params_sampling is not None:
        paramset, defaults = _modify_sampling_parameters(
            paramset, params_sampling,
            params_default=defaults, ret_defaults=True
        )

    params_measure = {}
    if degree is not None:
        params_measure.update({'ELL': degree})
    if binning is not None:
        params_measure.update({
            'bin_min': binning.bin_min,
            'bin_max': binning.bin_max,
            'num_bins': binning.num_bins,
        })

    paramset, defaults = _modify_measurement_parameters(
        paramset, params_measure,
        params_default=defaults, ret_defaults=True
    )

    # Implicitly validate parameters.
    paramset.update(type_kwargs)

    if defaults:
        warnings.warn(
            "The following parameter default values "
            f"are unchanged: {defaults}. "
            "Not all parameters are necessarily used."
        )

    return paramset


def _get_measurement_filename(paramset):
    """Get output measurement filename.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.

    Returns
    -------
    str
        Output measurement filename.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        two-point statistics.

    """
    multipole = paramset['degrees']['ELL']

    output_tag = paramset['tags']['output'] or ""

    if paramset['statistic_type'] == 'powspec':
        return "pk{:d}{}".format(multipole, output_tag)
    if paramset['statistic_type'] == '2pcf':
        return "xi{:d}{}".format(multipole, output_tag)
    if paramset['statistic_type'] == '2pcf-win':
        return "xiw{:d}{}".format(multipole, output_tag)

    raise ValueError(
        "`paramset` 'statistic_type' does not correspond to a "
        "recognised two-point statistic."
    )


def _print_measurement_header(paramset, norm_factor_part, norm_factor_mesh,
                              norm_factor_meshes):
    """Print two-point statistic measurement header including
    sampling parameters, normalisation factors and data table columns.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.
    norm_factor_part, norm_factor_mesh, norm_factor_meshes : float
        Normalisation factors.

    Returns
    -------
    text_header : str
        Measurement information as a header string.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        two-point statistics.

    """
    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
    if paramset['norm_convention'] == 'mesh-mixed':
        norm_factor = norm_factor_meshes

    if paramset['npoint'] != '2pt':
        raise ValueError(
            "`paramset` 'statistic_type' does not correspond to a "
            "recognised two-point statistic."
        )
    if paramset['space'] == 'fourier':
        datatab_colnames = [
            "k_cen", "k_eff", "nmodes",
            "Re{{pk{:d}_raw}}".format(paramset['degrees']['ELL']),
            "Im{{pk{:d}_raw}}".format(paramset['degrees']['ELL']),
            "Re{{pk{:d}_shot}}".format(paramset['degrees']['ELL']),
            "Im{{pk{:d}_shot}}".format(paramset['degrees']['ELL'])
        ]
    if paramset['space'] == 'config':
        datatab_colnames = [
            "r_cen", "r_eff", "npairs",
            "Re{{xi{:d}}}".format(paramset['degrees']['ELL']),
            "Im{{xi{:d}}}".format(paramset['degrees']['ELL'])
        ]

    text_lines = [
        "Box size: [{:.3f}, {:.3f}, {:.3f}]".format(
            *[paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
        ),
        "Box alignment: {}".format(paramset['alignment']),
        "Mesh number: [{:d}, {:d}, {:d}]".format(
            *[paramset['ngrid'][ax] for ax in ['x', 'y', 'z']]
        ),
        "Mesh assignment and interlacing: {}, {}".format(
            paramset['assignment'], paramset['interlace']
        ),
        "Normalisation factor: {:.9e} ({})".format(
            norm_factor, paramset['norm_convention']
        ),
        "Normalisation factor alternatives: "
        "{:.9e} (particle), {:.9e} (mesh), {:.9e} (mesh-mixed)".format(
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        ),
        ", ".join([
            "[{:d}] {}".format(colidx, colname)
            for colidx, colname in enumerate(datatab_colnames)
        ])
    ]

    text_header = "\n".join(text_lines)

    return text_header


def _assemble_measurement_datatab(measurements, paramset):
    """Assemble measurement data table.

    Parameters
    ----------
    measurements : dict
        Measurement results.
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.

    Returns
    -------
    datatab : :class:`numpy.ndarray`
        Column-major measurement data table.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        two-point statistics.

    """
    if paramset['npoint'] != '2pt':
        raise ValueError(
            "Measurement header being printed is for two-point statistics."
        )
    if paramset['space'] == 'fourier':
        datatab = np.transpose([
            measurements['kbin'], measurements['keff'], measurements['nmodes'],
            measurements['pk_raw'].real, measurements['pk_raw'].imag,
            measurements['pk_shot'].real, measurements['pk_shot'].imag,
        ])
    if paramset['space'] == 'config':
        datatab = np.transpose([
            measurements['rbin'], measurements['reff'], measurements['npairs'],
            measurements['xi'].real, measurements['xi'].imag,
        ])

    return datatab


# ========================================================================
# Survey statistics
# ========================================================================

def _compute_2pt_stats_survey_like(twopt_algofunc,
                                   catalogue_data, catalogue_rand,
                                   los_data=None, los_rand=None,
                                   paramset=None, params_sampling=None,
                                   degree=None, binning=None, types=None,
                                   save=False, logger=None):
    """Compute two-point statistics from survey-like data and random
    catalogues in the local plane-parallel approximation.

    Parameters
    ----------
    twopt_algofunc : callable
        Two-point statistic algorithmic function.
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), `degree`, `binning`
        and `params_sampling` must be provided.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    types : dict, optional
        `catalogue_type` and `statistic_type` (default is `None`).
        This should be set by the caller of this function.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` while `params_sampling`, `degree` or
        `binning` is also `None`.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=params_sampling,
        degree=degree, binning=binning, **types
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up lines of sight.
    if los_data is None:
        los_data = catalogue_data.compute_los()
    if los_rand is None:
        los_rand = catalogue_rand.compute_los()

    los_data = np.ascontiguousarray(los_data)
    los_rand = np.ascontiguousarray(los_rand)

    if logger:
        logger.info("Lines of sight have been initialised.")

    # Set up box alignment.
    if paramset['alignment'] == 'centre':
        catalogue_data.centre(
            [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
            catalogue_ref=catalogue_rand
        )
    if paramset['alignment'] == 'pad':
        if paramset['padscale'] == 'box':
            kwargs = {'boxsize_pad': paramset['padfactor']}
        if paramset['padscale'] == 'grid':
            kwargs = {
                'ngrid': [paramset['ngrid'][ax] for ax in ['x', 'y', 'z']],
                'ngrid_pad': paramset['padfactor']
            }
        catalogue_data.pad(
            [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
            catalogue_ref=catalogue_rand, **kwargs
        )

    if logger:
        logger.info("Catalogues have been aligned.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    if logger:
        logger.info(
            "Preparing catalogue for clustering algorithm...",
            cpp_state='start'
        )

    particles_data = \
        catalogue_data._convert_to_cpp_catalogue(verbose=paramset['verbose'])
    particles_rand = \
        catalogue_rand._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    if logger:
        logger.info(
            "... prepared catalogue for clustering algorithm.",
            cpp_state='end'
        )

    # Set up constants.
    alpha = catalogue_data.wstotal / catalogue_rand.wstotal

    if logger:
        logger.info("Alpha contrast: %.6e.", alpha)

    norm_factor_part = _calc_powspec_normalisation_from_particles(
        particles_rand, alpha
    )
    norm_factor_mesh = _calc_powspec_normalisation_from_mesh(
        particles_rand, paramset, alpha
    )
    norm_factor_meshes = _calc_powspec_normalisation_from_meshes(
        particles_data, particles_rand, paramset, alpha,
        padding=PADDING, cellsize=CELLSIZE, assignment=ASSIGNMENT
    )

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed)."
        )
    if paramset['norm_convention'] == 'mesh-mixed':
        norm_factor = norm_factor_meshes
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; used)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info("Measuring clustering statistics...", cpp_state='start')

    results = twopt_algofunc(
        particles_data, particles_rand, los_data, los_rand,
        paramset, binning, norm_factor
    )

    if logger:
        logger.info("... measured clustering statistics.", cpp_state='end')

    if save:
        odirpath = paramset['directories']['measurements'] or ""
        header = "\n".join([
            catalogue_data.write_attrs_as_header(catalogue_ref=catalogue_rand),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                ['%.9e'] * 2 + ['%10d'] + ['% .9e'] * (datatab.shape[-1] - 3)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results


def compute_powspec(catalogue_data, catalogue_rand,
                    los_data=None, los_rand=None,
                    degree=None, binning=None, sampling_params=None,
                    paramset=None,
                    save=False, logger=None):
    """Compute power spectrum from survey-like data and random catalogues
    in the local plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degree`, `binning` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` but `degree`, `binning` or
        `sampling_params` is also `None`.

    Examples
    --------
    Specify line-of-sight vectors explicitly.

    >>> results = compute_powspec(
    ...     catalogue_data, catalogue_rand,
    ...     los_data=np.ones((1e3, 3)), los_rand=np.ones((1e4, 3)),
    ...     paramset=None
    ... )

    Specify multipole `degree` 2, customised
    :class:`~triumvirate.dataobjs.Binning` object ``binning``.
    Whether `paramset` provided or not, relevant parameters are overriden
    by the supplied keyword arguments.

    >>> results = compute_powspec(
    ...     catalogue_data, catalogue_rand,
    ...     degree=2,
    ...     binning=binning,
    ...     sampling_params={
    ...         'boxsize': [1000., 1500., 1000.],
    ...         'ngrid': [256, 256, 256],
    ...         # 'boxalign' at default initial value in `ParameterSet`
    ...         # 'assignment' at default initial value in `ParameterSet`
    ...     }
    ... )

    """
    # if logger:
    #     logger.info(
    #         "Measuring power spectrum from paired survey-type catalogues...",
    #         cpp_state='start'
    #     )

    results = _compute_2pt_stats_survey_like(
        _compute_powspec,
        catalogue_data, catalogue_rand,
        los_data=los_data, los_rand=los_rand,
        paramset=paramset, params_sampling=sampling_params,
        degree=degree, binning=binning,
        types={'catalogue_type': 'survey', 'statistic_type': 'powspec'},
        save=save, logger=logger
    )

    # if logger:
    #     logger.info(
    #         "... measured power spectrum from paired survey-type catalogues.",  # noqa: E501
    #         cpp_state='end'
    #     )

    return results


def compute_corrfunc(catalogue_data, catalogue_rand,
                     los_data=None, los_rand=None,
                     degree=None, binning=None, sampling_params=None,
                     paramset=None,
                     save=False, logger=None):
    """Compute correlation function from survey-like data and random
    catalogues in the local plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used
        in lieu of `degree`, `binning` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` but `degree`, `binning` or
        `sampling_params` is also `None`.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.twopt.compute_powspec`.

    """
    # if logger:
    #     logger.info(
    #         "Measuring two-point correlation function "
    #         "from paired survey-type catalogues...",
    #         cpp_state='start'
    #     )

    results = _compute_2pt_stats_survey_like(
        _compute_corrfunc,
        catalogue_data, catalogue_rand,
        los_data=los_data, los_rand=los_rand,
        paramset=paramset, params_sampling=sampling_params,
        degree=degree, binning=binning,
        types={'catalogue_type': 'survey', 'statistic_type': '2pcf'},
        save=save, logger=logger
    )

    # if logger:
    #     logger.info(
    #         "... measured two-point correlation function "
    #         "from paired survey-type catalogues.",
    #         cpp_state='end'
    #     )

    return results


# ========================================================================
# Simulation statistics
# ========================================================================

def _compute_2pt_stats_sim_like(twopt_algofunc, catalogue_data,
                                paramset=None, params_sampling=None,
                                degree=None, binning=None, types=None,
                                save=False, logger=None):
    """Compute two-point statistics from a simulation-box catalogue
    in the global plane-parallel approximation.

    Parameters
    ----------
    twopt_algofunc : callable
        Two-point statistic algorithmic function.
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), `degree`, `binning`
        and `sampling_params` must be provided.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    types : dict, optional
        `catalogue_type` and `statistic_type` (default is `None`).
        This should be set by the caller of this function.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` while `params_sampling`, `degree` or
        `binning` is also `None`.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=params_sampling,
        degree=degree, binning=binning, **types
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up box alignment.
    catalogue_data.periodise(
        [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
    )

    if logger:
        logger.info("Catalogue box has been periodised.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    if not catalogue_data['nz'].any():
        catalogue_data.compute_mean_density(
            boxsize=list(paramset['boxsize'].values())
        )
        if logger:
            logger.info(
                "Inserted missing 'nz' field "
                "based on particle count and boxsize."
            )

    if logger:
        logger.info(
            "Preparing catalogue for clustering algorithm...",
            cpp_state='start'
        )

    particles_data =  \
        catalogue_data._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    if logger:
        logger.info(
            "... prepared catalogue for clustering algorithm.",
            cpp_state='end'
        )

    # Set up constants.
    norm_factor_part = _calc_powspec_normalisation_from_particles(
        particles_data, alpha=1.
    )
    norm_factor_mesh = _calc_powspec_normalisation_from_mesh(
        particles_data, paramset, alpha=1.
    )
    norm_factor_meshes = 0.

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info("Measuring clustering statistics...", cpp_state='start')

    results = twopt_algofunc(particles_data, paramset, binning, norm_factor)

    if logger:
        logger.info("... measured clustering statistics.", cpp_state='end')

    if save:
        odirpath = paramset['directories']['measurements'] or ""
        header = "\n".join([
            catalogue_data.write_attrs_as_header(),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                ['%.9e'] * 2 + ['%10d'] + ['% .9e'] * (datatab.shape[-1] - 3)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results


def compute_powspec_in_gpp_box(catalogue_data,
                               degree=None, binning=None, sampling_params=None,
                               paramset=None,
                               save=False, logger=None):
    """Compute power spectrum from a simulation-box catalogue
    in the global plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used
        in lieu of `degree`, `binning` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` but `degree`, `binning` or
        `sampling_params` is also `None`.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.twopt.compute_powspec`
    (though without line-of-sight arguments).

    """
    # if logger:
    #     logger.info(
    #         "Measuring power spectrum from a simulation-box catalogue "
    #         "in the global plane-parallel approximation...",
    #         cpp_state='start'
    #     )

    results = _compute_2pt_stats_sim_like(
        _compute_powspec_in_gpp_box, catalogue_data,
        paramset=paramset, params_sampling=sampling_params,
        degree=degree, binning=binning,
        types={'catalogue_type': 'sim', 'statistic_type': 'powspec'},
        save=save, logger=logger
    )

    # if logger:
    #     logger.info(
    #         "... measured power spectrum from a simulation-box catalogue "
    #         "in the global plane-parallel approximation.",
    #         cpp_state='end'
    #     )

    return results


def compute_corrfunc_in_gpp_box(catalogue_data,
                                degree=None, binning=None,
                                sampling_params=None,
                                paramset=None,
                                save=False, logger=None):
    """Compute correlation function from a simulation-box catalogue
    in the global plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used
        in lieu of `degree`, `binning` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` but `degree`, `binning` or
        `sampling_params` is also `None`.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.twopt.compute_powspec`
    (though without line-of-sight arguments).

    """
    # if logger:
    #     logger.info(
    #         "Measuring two-point correlation function "
    #         "from a simulation-box catalogue "
    #         "in the global plane-parallel approximation...",
    #         cpp_state='start'
    #     )

    results = _compute_2pt_stats_sim_like(
        _compute_corrfunc_in_gpp_box, catalogue_data,
        paramset=paramset, params_sampling=sampling_params,
        degree=degree, binning=binning,
        types={'catalogue_type': 'sim', 'statistic_type': '2pcf'},
        save=save, logger=logger
    )

    # if logger:
    #     logger.info(
    #         "... measured two-point correlation function "
    #         "from a simulation-box catalogue "
    #         "in the global plane-parallel approximation.",
    #         cpp_state='end'
    #     )

    return results


# ========================================================================
# Window statistics
# ========================================================================

def compute_corrfunc_window(catalogue_rand, los_rand=None,
                            degree=None, binning=None, sampling_params=None,
                            paramset=None,
                            save=False, logger=None):
    """Compute correlation function window from a random catalogue.

    Parameters
    ----------
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degree : int, optional
        Multipole degree.  If not `None` (default), this will override
        ``paramset['degrees']['ELL']``.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default), this is
        constructed from `paramset`.
    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'boxalign': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and one and only one of the following when 'boxalign' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        This will override corresponding entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used
        in lieu of `degree`, `binning` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format. The save path is determined from `paramset`
        (if unset, a default file in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    Raises
    ------
    ValueError
        When `paramset` is `None` but `degree`, `binning` or
        `sampling_params` is also `None`.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.twopt.compute_powspec`.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=sampling_params,
        degree=degree, binning=binning,
        catalogue_type='random', statistic_type='2pcf-win'
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up lines of sight.
    if los_rand is None:
        los_rand = catalogue_rand.compute_los()
    los_rand = np.ascontiguousarray(los_rand)

    if logger:
        logger.info("Lines of sight have been initialised.")

    # Set up box alignment.
    catalogue_rand.centre(
        [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
    )

    if logger:
        logger.info("Catalogues have been aligned.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    particles_rand =  \
        catalogue_rand._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    # Set up constants.
    norm_factor_part = _calc_powspec_normalisation_from_particles(
        particles_rand, alpha=1.
    )
    norm_factor_mesh = _calc_powspec_normalisation_from_mesh(
        particles_rand, paramset, alpha=1.
    )
    norm_factor_meshes = 0.

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info(
            "Measuring window function statistics...", cpp_state='start'
        )

    results = _compute_corrfunc_window(
        particles_rand, los_rand, paramset, binning,
        alpha=1., norm_factor=norm_factor
    )

    if logger:
        logger.info(
            "... measured window function statistics.", cpp_state='end'
        )

    if save:
        odirpath = paramset['directories']['measurements']
        header = "\n".join([
            catalogue_rand.write_attrs_as_header(),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                ['%.9e'] * 2 + ['%10d'] + ['% .9e'] * (datatab.shape[-1] - 3)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results



"""Test :mod:`~triumvirate.dataobjs`.

"""
import numpy as np
import pytest

from triumvirate.dataobjs import Binning


@pytest.fixture
def default_binning(valid_paramset):
    binning = Binning.from_parameter_set(valid_paramset)
    return binning


@pytest.mark.parametrize(
    "space, scheme, bin_min, bin_max, num_bins",
    [
        ('config', 'logpad', 1.e-1, 1.e4, 20),
        ('fourier', 'lin', None, None, None),
    ]
)
def test_Binning___cinit__(space, scheme, bin_min, bin_max, num_bins):
    Binning(space, scheme, bin_min=bin_min, bin_max=bin_max, num_bins=num_bins)


def test_Binning_from_parameter_set(valid_paramset):
    Binning.from_parameter_set(valid_paramset)


@pytest.mark.parametrize(
    "bin_min, bin_max, num_bins",
    [
        (1.e-1, 1.e4, 20),
        (5.e-3, 1.e-1+5.e-3, 10),
    ]
)
def test_Binning_set_bins(bin_min, bin_max, num_bins, default_binning):
    default_binning.set_bins(bin_min, bin_max, num_bins)
    assert np.allclose(
        default_binning.bin_edges,
        np.linspace(bin_min, bin_max, num=num_bins+1)
    ), "Bin resetting is not correct."


@pytest.mark.parametrize(
    "boxsize, ngrid",
    [
        (1000., [16,]*3),  # noqa: E231
        ([1000.,]*3, 16),  # noqa: E231
    ]
)
def test_Binning_set_grid_based_bins(boxsize, ngrid, default_binning):

    default_binning.set_grid_based_bins(boxsize, ngrid)

    bin_width = 2*np.pi / np.max(boxsize)

    bin_min = 0.
    bin_max = np.pi * np.min(ngrid) / np.max(boxsize) + bin_width/2.
    num_bins = np.min(ngrid) // 2

    assert np.allclose(
        default_binning.bin_edges,
        np.linspace(bin_min, bin_max, num=num_bins+1)
    ), "Grid-based bin resetting is not correct."


@pytest.mark.parametrize(
    "bin_edges",
    [
        np.sort(np.random.rand(10)),
    ]
)
def test_Binning_set_custom_bins(bin_edges, default_binning):

    default_binning.set_custom_bins(bin_edges)

    bin_centres = (bin_edges[1:] + bin_edges[:-1]) / 2
    bin_widths = bin_edges[1:] - bin_edges[:-1]

    assert np.allclose(
        default_binning.bin_edges,
        bin_edges
    ), "Customised bin resetting is not correct in bin edges."
    assert np.allclose(
        default_binning.bin_centres,
        bin_centres
    ), "Customised bin resetting is not correct in bin centres."
    assert np.allclose(
        default_binning.bin_widths,
        bin_widths
    ), "Customised bin resetting is not correct in bin widths."
